/*--------------------------------------------------------------------
 *    Scripps Institution of Oceanography: @(#)pluswiggle.c  1.0  5/18/94
 *    based on:
 *
 *    The GMT-system:   pswiggle.c  2.8  7/22/93
 *
 *    Copyright (c) 1991 by P. Wessel and W. H. F. Smith
 *    See README file for copying and redistribution conditions.
 *
 *--------------------------------------------------------------------*/
/*
 * pluswiggle reads x,y,z from stdin and plots a wiggleplot using the
 * specified map projection. If the distance between 2 consecutive points
 * exceeds the value of dist_gap, a data gap is assumed.  The user may
 * select a preferred direction where s/he wants the positive anomalies
 * to be pointing towards.  Positive normal vectors to the trackline
 * will then always be withing +- 90 degrees of this direction.
 * Separate colors may be specified for positive anomaly, outline of
 * anomaly, and trackline.  Plotting of the outline and track is optional.
 *
 * Author:      Paul Wessel
 * Date:        24-APR-1991
 * Version:     2.0, based on old v1.x
 * Revised:     5/11/94 by Ed Stickgold for SIO
 *              Name of revised version changed to "pluswiggle"
 *              8/19/04 by Ed Stickgold for SIO
 *              Minor changes for GMTPlus to work with GMT 3.4.3
 *              10/26/04 by Ed Stickgold for SIO
 *              added call to GMT_put_history to write .gmtcommands file
 * Version:     1.0 (of revised version named pluswiggle)
 *
 */
 
#include <gmt.h>
#ifdef __APPLE__
#  include <limits.h>
#define MAXDOUBLE __DBL_MAX__
#else
#  include <malloc.h>
#  include <values.h>
#endif




/* configurable defines */
#define MAX_POINTS         400          /* Max points to store up before sending to postscript file */
#define FLIP_DELAY         5.0          /* Number of degrees to delay Z-axis flip */
#define INCLUSION_SPREAD   45.0         /* Default number of degrees included for restricted ship headings */
#define SMOOTH_DIST_MAX    5            /* Default max distance (km) for points used to smooth headings option */
#define SMOOTH_POINTS      2            /* Default num points to use on either side for smooth headings option */

/* Z-AXIS FLIP  CONSTANTS */
#define FLIP_STATE_NO_FLIP 0            /* Z-axis is not currently flipped */
#define FLIP_STATE_FLIP    180          /* Z-axis is currently flipped */
#define NO_DELAYED_FLIP    FALSE        /* Not using delayed Z-axis flips */
#define DELAYED_FLIP       TRUE         /* Using delayed Z-axis flips */
#define FLIP_DELAY         5.0          /* Number of degrees to delay Z-axis flip */

/* misc globals (non-static correct?) */
double *xx, *yy, *zz, *lon, *lat, *z;
double *hh = NULL; /* smoothed angular heading of projected track */
double *dd = NULL; /* distance along the actual track */

/* SMOOTH HEADINGS VARIABLES */
static BOOLEAN smooth_headings = FALSE; /* smooth projected track direction for z-axis determination */
static float   smooth_distance_max = 1000.0 * SMOOTH_DIST_MAX; /* (convert to meters) */
static long    smooth_points = SMOOTH_POINTS;

/* Z-AXIS FLIP VARIABLES */
static double start_az, stop_az; /* z-axis flip points when not using a delayed flip */
static double no_flip_start_az, no_flip_stop_az; /* delayed flip points while in unflipped state */
static double flip_start_az, flip_stop_az; /* delayed flip points while in flipped state */
static long   current_flip_state = FLIP_STATE_NO_FLIP; /* initial flipped state */


/* int plot_wiggle (x, y, z, np, zscale, fill, pen_o, pen_t, paint_wiggle, outline, track) */
/* above line changed to line below 4/13/94 - EAS */


int plot_wiggle (x, y, z, np, zscale, fill, pen_o, pen_t, paint_wiggle, outline, track, use_fixed_az, fixed_az, delayed_flip)

double x[], y[], z[], zscale;
int np;
struct GMT_FILL *fill;
struct GMT_PEN *pen_o, *pen_t;
BOOLEAN paint_wiggle, outline, track, use_fixed_az; 
double fixed_az;                /* angle to use for z axis if z axis has been fixed */
BOOLEAN delayed_flip;
{
	double dx, dy, x_inc, y_inc;
	static double az = 0.0;
	int n, i;
	
	if (paint_wiggle || outline) {
		if (!GMT_n_alloc) GMT_get_plot_array ();
		GMT_x_plot[0] = x[0];
		GMT_y_plot[0] = y[0];
		n = 1;
		if (use_fixed_az) az = (-fixed_az) * D2R; /* added 4/13/94 - EAS */
		for (i = 0; i < np; i++) {
			if (i < np-1) {
				if (!use_fixed_az) {    /* if condition added 4/13/94 - EAS */
					if (smooth_headings) 
						az = hh[i]; /* guaranteed to be between -180 and +180 */
					else {
						dx = x[i+1] - x[i];
						dy = y[i+1] - y[i];
						if (!(dx == 0.0 && dy == 0.0)) 
							az = d_atan2 (dy, dx) * R2D - 360.0;
						/* else az will equal previous value by default since az is static */
					}
					/* note: (d_atan2() * R2D) will always return a number between -180. and +180. */
					/*       so if az is set here, it will be between -540. and -180.0 */
					if (!delayed_flip) {
						while (az < start_az) az += 360.0;
						if (az > stop_az) az -= 180.0;
					}
					else { /* delayed flipping . . . */
						if (current_flip_state == FLIP_STATE_FLIP) {
							while (az < flip_start_az) az += 360.0;
							if (az > flip_stop_az) {
								az -= 180.0;
								current_flip_state = FLIP_STATE_NO_FLIP;
							}
						}
						else {
							while (az < no_flip_start_az) az += 360.0;
							if (az > no_flip_stop_az) {
								az -= 180.0;
								current_flip_state = FLIP_STATE_FLIP;
							}
						}
					}
					az *= D2R; /* D2R = Degrees to Radians */
				}
			}
			if (fabs (z[i]) > 0.0) {
				x_inc = -zscale * z[i] * sin (az);
				y_inc = zscale * z[i]* cos (az);
			}
			else
				x_inc = y_inc = 0.0;
				
			GMT_x_plot[n] = x[i] + x_inc;
			GMT_y_plot[n] = y[i] + y_inc;
			n++;
			if (n == GMT_n_alloc) GMT_get_plot_array ();
		}
		GMT_x_plot[n] = x[np-1];
		GMT_y_plot[n] = y[np-1];
		n++;
		
		if (paint_wiggle) {
			for (i = np - 2; i >= 0; i--, n++) {    /* Go back to 1st point along track */
				if (n == GMT_n_alloc) GMT_get_plot_array ();
				GMT_x_plot[n] = x[i];
				GMT_y_plot[n] = y[i];
			}
		}
#ifdef GMT_LEVEL  /* GMT 4 I hope! */
		fprintf (stderr, "FIX ERROR!!!! SORRY - need to port call to GMT_2D_to_3D to GMT 4\n");
#else
		if (project_info.three_D) GMT_2D_to_3D (GMT_x_plot, GMT_y_plot, n);
#endif

	}
	
	
	if (paint_wiggle) { /* First shade wiggles */
		printf ("%% Positive wiggle\n");
		if (fill->use_pattern)
			ps_imagefill (GMT_x_plot, GMT_y_plot, n, fill->pattern_no, fill->pattern, fill->inverse, fill->dpi, FALSE, fill->colorize, fill->f_rgb, fill->b_rgb);
		else
			ps_polygon (GMT_x_plot, GMT_y_plot, n, fill->rgb, FALSE);
	}
	
	if (outline) { /* Then draw wiggle outline */
		printf ("%% Wiggle line\n");
		ps_setpaint (pen_o->rgb);
		ps_setline (pen_o->width);
		/* jjb: assume we can "SPLIT" line into sevral sections */
		ps_line (&GMT_x_plot[1], &GMT_y_plot[1], np, 3, FALSE, TRUE);
	}
	
	if (track) {    /* Finally draw track line */
#ifdef GMT_LEVEL /* Please be GMT 4 */
		fprintf (stderr, "FIX ERROR!!!! SORRY - need to port call to GMT_2D_to_3D to GMT 4\n");
#else
		if (project_info.three_D) GMT_2D_to_3D (x, y, np);
#endif
		printf ("%% Track line\n");
		ps_setline (pen_t->width);
		ps_setpaint (pen_t->rgb);
		/* jjb: assume we can "SPLIT" line into sevral sections */
		ps_line (x, y, np, 3, FALSE, TRUE);
	}
}

/* following function should be embedded where called (for speed) */

double angular_diff(double nearby_heading, double current_heading)
{
  double retval;
  retval = nearby_heading-current_heading;
  while (retval < -180.0) retval += 360.0;
  while (retval > 180.0)  retval -= 360.0;
  return retval;
}

int 
main (int argc, char *argv[])
{
	int i, j, k, n, ix, iy, n_files = 0, fno, n_alloc = GMT_CHUNK, n_args;
	char *ptr, *slashptr;
	BOOLEAN use_fixed_az = FALSE, delayed_flip = NO_DELAYED_FLIP;
	double  fixed_az;  /* angle to use for z axis if z axis has been fixed */

	BOOLEAN excluded_headings = FALSE;
	BOOLEAN ship_heading_ok = TRUE;
	BOOLEAN around_zero_degs;
	double  ship_heading;
	double  ship_heading_c;
	double  excluded_headings_center;
	double  included_headings_min;
	double  included_headings_max;

	BOOLEAN z_axis_clip = FALSE;
	BOOLEAN z_axis_ok = TRUE;
	double  z_axis_min;
	double  z_axis_max;
	double  tmp;

	BOOLEAN xy_clip = FALSE;
	BOOLEAN xy_clip_ok = TRUE;
	double  x_out, y_out;
	
	BOOLEAN error = FALSE, nofile = TRUE, outline = FALSE, track = FALSE, paint_wiggle;
	BOOLEAN multi_segments = FALSE, do_fill = FALSE, negative = FALSE, done;
	
	BOOLEAN use_zoffset = FALSE;
	double  zoffset = 0.0;

	double west = 0.0, east = 360.0, south = 0.0, north = 0.0, x_2, y_2;
	double zscale = 0.0, dx, dy, dz, ds, dist_gap = MAXDOUBLE, xy[2], pref_dir = 0.0;
	double x1, y1, x2, y2;

	double *smooth_cache_hh = NULL;
	double az_smoothing;
	int az_smoothing_num;
	double dist_tot;
	int skipped_headings = 0;
	int last_heading;
	
	char line[512], *more, EOF_flag = '>';
	
	FILE *fp = NULL;
	
	struct GMT_FILL fill;
	struct GMT_PEN pen_o, pen_t;
	
	/* initialize defaults and retrieve selected options from .gmtcommands */
	GMT_begin (argc, argv);
	
	fill.rgb[0] = pen_o.rgb[0] = pen_t.rgb[0] = gmtdefs.basemap_frame_rgb[0];
	fill.rgb[1] = pen_o.rgb[1] = pen_t.rgb[1] = gmtdefs.basemap_frame_rgb[1];
	fill.rgb[2] = pen_o.rgb[2] = pen_t.rgb[2] = gmtdefs.basemap_frame_rgb[2];
	fill.use_pattern = FALSE;
	pen_o.width = pen_t.width = 1;

	/* set parameters */
	for (i = 1; !error && i < argc; i++) {
		if (argv[i][0] == '-') {
			switch(argv[i][1]) {
		
				/* Common parameters */
			
				case 'B':
				case 'H':
				case 'J':
				case 'K':
				case 'O':
				case 'P':
				case 'R':
				case 'U':
				case 'V':
				case 'X':
				case 'x':
				case 'Y':
				case 'y':
				case '#':
				case ':':
				case '\0':
					error += GMT_get_common_args (argv[i], &west, &east, &south, &north);
					break;
				
				/* Supplemental parameters */
			
				case 'A':
					pref_dir = atof (&argv[i][2]);
					break;
				case 'D':
					dist_gap = atof (&argv[i][2]) * 1000.0;
					break;
				case 'E':
					sscanf (&argv[i][2], "%lf/%lf", &z_project.view_azimuth, &z_project.view_elevation);
					break;
				case 'F':
					sscanf (&argv[i][2], "%d/%d/%d",&gmtdefs.basemap_frame_rgb[0],
						&gmtdefs.basemap_frame_rgb[1], &gmtdefs.basemap_frame_rgb[2]);
					break;
				case 'G':
					GMT_getfill (&argv[i][2], &fill);
					do_fill = TRUE;
					break;
				case 'M':
					if (argv[i][2]) EOF_flag = argv[i][2];
					multi_segments = TRUE;
					break;
				case 'N':
					negative = TRUE;
					break;
				case 'S':       /* suppress data points outside of output plot */
					xy_clip = TRUE;
					break;
				case 'T':
					switch(argv[i][2]) {
						case 'H': /* track hysteresis */
							delayed_flip = DELAYED_FLIP;
							break;
						case 'I': /* track include */
							if ((slashptr = strchr(argv[i]+3, '/')) == NULL) { /* ctr position provided */
								excluded_headings_center = strtod(argv[i]+3, &ptr);
								if (excluded_headings = (ptr != (argv[i]+3))) { /* number ok? */
									included_headings_min = excluded_headings_center -
												INCLUSION_SPREAD/2;
									included_headings_max = excluded_headings_center +
												INCLUSION_SPREAD/2;
								}
							}
							else { /* min & max included ship headings provided explicitly */
								included_headings_min = strtod(argv[i]+3, &ptr);
								if (excluded_headings = (ptr != (argv[i]+3))) { /* number ok? */
									included_headings_max = strtod(slashptr+1, &ptr);
									excluded_headings = (ptr != (slashptr+1)); /* number ok? */
								}
							}
							if (excluded_headings) { /* if parameters are acceptable */
								/* normalize & check data validity */
								/* min/max range from 0.0 to 360.0 */
								while (included_headings_min > 360.0) included_headings_min -= 360.0;
								while (included_headings_max > 360.0) included_headings_max -= 360.0;
								while (included_headings_min < 0.0) included_headings_min += 360.0;
								while (included_headings_max < 0.0) included_headings_max += 360.0;
								if (fabs((included_headings_min - included_headings_max)) < .0001) {
									/* too close! - assume they are the same */
									excluded_headings = FALSE;
									error = TRUE;
								}
								around_zero_degs = (included_headings_max < included_headings_min);
							}
							else error = TRUE; /* else indicate an error */
							break;
						case 'S': /* smooth projected track heading (for z-axis determination only) */
							smooth_headings = TRUE;
							if (ptr = strchr(argv[i]+3, 'p'))  /* number of points in each direction */
								if ((smooth_points = atol(ptr+1)) == 0) error = TRUE;
							if (ptr = strchr(argv[i]+3, 'd'))  /* distance in each direction */
								if ((smooth_distance_max = 1000.0 * atof(ptr+1)) == 0.0) error = TRUE;
							break;
						default:
							GMT_getpen (&argv[i][2], &pen_t);
							track = TRUE;
							break;
					}       
					break;
				case 'W':
					switch(argv[i][2]) {
						case 'A':       /* set fixed direction z-axis */
							fixed_az = strtod(argv[i]+3, &ptr);
							use_fixed_az = (ptr != (argv[i]+3));
							break;
						case 'C':       /* clip z axis */
							sscanf (&argv[i][3], "%lf/%lf", &z_axis_min, &z_axis_max);
							if (z_axis_min != z_axis_max) {
								if (z_axis_min > z_axis_max) {
									tmp = z_axis_max;
									z_axis_max = z_axis_min;
									z_axis_min = tmp;
								}
								z_axis_clip = TRUE;
							}
							else z_axis_clip = FALSE;
							break;
						default:
							GMT_getpen (&argv[i][2], &pen_o);
							outline = TRUE;
							break;
					}
					break;
				case 'Z':
					if ((slashptr = strchr(argv[i]+2, '/')) == NULL)  /* just scale provided */
						zscale = atof (&argv[i][2]);
					else {
						sscanf (&argv[i][2], "%lf/%lf", &zscale, &zoffset);
						use_zoffset = TRUE;
					}
					break;
					
				/* Illegal options */
			
				default:
					error = TRUE;
					break;
			}
		}
		else
			n_files++;
	}


	/* check region in -R for valid range */
	error += GMT_check_region (west, east, south, north);
	if (zscale == 0.0) error = TRUE;
	/* check rgbs for valid range */
	error += GMT_check_rgb (pen_o.rgb)
	      + GMT_check_rgb (fill.rgb)
	      + GMT_check_rgb (pen_t.rgb)
	      + GMT_check_rgb (gmtdefs.basemap_frame_rgb);
	      
	/* exit(1); */
	if (argc == 1 || error) {
		fprintf (stderr,"pluswiggle %s - Plot xyz-series along tracks\n\n", GMT_VERSION);
		fprintf (stderr, "usage: pluswiggle <xyz-files> -J<params> -R<west/east/south/north> -Z<scale[/offset]>\n");
		fprintf (stderr, "	[-A<direction>] [-B<tickinfo>] [-D<gap>] [-Eaz/el] [-F<r/g/b>] [-G<fill>] [-H]\n");
		fprintf (stderr, "	[-K] [-M[<flag>]] [-N] [-O] [-P] [-S] [-T<trackpen>] [-TH] [-TI<center>|<start/stop>]\n");
		fprintf (stderr, "	[-TS[d<distance>][p<points]] [-U] [-V] [-W<outlinepen>] [-WA<z_ctr-azimuth>]\n");
		fprintf (stderr, "	[-WC<z_low/z_high>] [-X<x_shift>] [-Y<y_shift>] [-#<ncopies>] [-:]\n");

		if (GMT_quick) exit (-1);
		
		fprintf (stderr, "	<xyz_files> is one or more files.  If none, read standard input\n");
		GMT_explain_option ('J');
		GMT_explain_option ('R');
		fprintf (stderr, "	-Z gives the wiggle scale in data-units pr. inch. If a second, optional,\n");
		fprintf (stderr, "	   argument is supplied (<scale/offset>), the argument is added to each\n");
		fprintf (stderr, "	   data value.\n");
		fprintf (stderr, "\n\tOPTIONS:\n");
		fprintf (stderr, "	-A set azimuth for preferred positive wiggle direction [0.0]\n");
		GMT_explain_option ('B');
		fprintf (stderr, "	-D means there is a datagap if 2 points are more than <gap> distance units apart\n");
		fprintf (stderr, "	   gap is in km for map projections, user-units otherwise\n");
		fprintf (stderr, "	-E set azimuth and elevation of viewpoint for 3-D perspective [180/90]\n");
		fprintf (stderr, "	-F Set color used for Frame and anotation [%d/%d/%d]\n",
			gmtdefs.basemap_frame_rgb[0], gmtdefs.basemap_frame_rgb[1], gmtdefs.basemap_frame_rgb[2]);
		fprintf (stderr, "	-G Specify color or pattern for positive areas. fill can be either\n");
		fprintf (stderr, "	   1) <r/g/b> (each 0-255) for color or <gray> (0-255) for gray-shade [0]\n");
		fprintf (stderr, "	   2) p[or P]<iconsize>/<pattern> for predefined patterns (0-31).\n");
		GMT_explain_option ('H');
		GMT_explain_option ('K');
		fprintf (stderr, "	-M Input files each consist of multiple segments separated by one record\n");
		fprintf (stderr, "	   whose first character is <flag> [>]\n");
		fprintf (stderr, "	-N Fill negative wiggles instead [Default is positive]\n");
		GMT_explain_option ('O');
		GMT_explain_option ('P');
		fprintf (stderr, "	-S Suppress data points outside of output plot. [Default is no suppression]\n");
		fprintf (stderr, "	-T specifies track pen attributes. [Default is no track]\n");
		fprintf (stderr, "	-TH Use hysteresis when wiggle is flipping back and forth about the preferred\n");
		fprintf (stderr, "	   positive wiggle direction. [Default is no hysteresis]\n");
		fprintf (stderr, "	   This means that the projected track angle must overshoot the flip angle\n");
		fprintf (stderr, "	   by %1.0f degrees before the z-axis flips direction (180 degrees).\n", FLIP_DELAY);
		fprintf (stderr, "	-TI Set to include points only within specified ship headings <ctrheading>\n");
		fprintf (stderr, "	   or <startheading/stopheading>. [Default includes all ship headings.]\n");
		fprintf (stderr, "	   NOTE: Ship heading must be provided as fourth data element from gmtlist.\n");
		fprintf (stderr, "	   If one heading is given,   then include that heading +- %4.1f degs.\n", 
			 INCLUSION_SPREAD/2.0);
		fprintf (stderr, "	   If two headings are given, then include that range of headings.\n");
		fprintf (stderr, "	-TS Smooth projected track directions for computing z-axis orientation.\n");
		fprintf (stderr, "	   Use a default of averaging no more than the 2 closest points in either\n");
		fprintf (stderr, "	   direction, so long as the points are within 5 km from the current ship\n");
		fprintf (stderr, "	   position. [Default without this option is no smoothing]\n");
		fprintf (stderr, "	   Max distance or points may be changed by providing one or both of the \n");
		fprintf (stderr, "	   following parameters (in either order:\n");
		fprintf (stderr, "	   d<MaxDistanceKm> specifies max distance from current ship position.\n");
		fprintf (stderr, "	   p<MaxNumPoints> specifies max number of points in either direction.\n");
		GMT_explain_option ('U');
		GMT_explain_option ('V');
		fprintf (stderr, "	-W specifies outline pen attributes. [Default is no outline]\n");
		fprintf (stderr, "	-WA set a fixed angle for z-axis. [Default is normal to ship heading]\n");
		fprintf (stderr, "	-WC remove points where z-value is outside of range <lowvalue/highvalue>.\n");
		GMT_explain_option ('X');
		GMT_explain_option ('#');
		GMT_explain_option (':');
		GMT_explain_option ('.');
		exit (-1);
	}

	GMT_put_history (argc,argv);	/* Update .gmtcommands */
	
	/* Initialize mapping system (see gmt_map.c) */
	GMT_map_setup (west, east, south, north);

	/* initialize parameters/open plotfile etc. */
	/* GMT_LEVEL defined in gmt.h - (/sw/include with fink) */
#ifdef GMT_LEVEL /* GMT 4 Please */
	fprintf (stderr, "FIX ERROR!!!! SORRY - need to port call to ps_plotinit to GMT 4\n");
#else /* I hope this is GMT 3 */
	ps_plotinit (NULL, gmtdefs.overlay, gmtdefs.page_orientation, gmtdefs.x_origin, gmtdefs.y_origin,
		gmtdefs.global_x_scale, gmtdefs.global_y_scale, gmtdefs.n_copies,
		gmtdefs.dpi, gmtdefs.measure_unit, gmtdefs.paper_width, gmtdefs.page_rgb, GMT_epsinfo (argv[0]));
#endif


	/* puts the command line into the PostScript file as comments */
	GMT_echo_command (argc, argv);
	/* plot UNIX time stamp with optional string */
	/* next line commented out because the variables are not defined. AMG, SIO, 6/5/96*/
	/*if (gmtdefs.unix_time) timestamp (TIME_STAMP_X, TIME_STAMP_Y, argc, argv);*/
	/* Translate and rotate the coordinate system */
	if (project_info.three_D) ps_transrotate (-z_project.xmin, -z_project.ymin, 0.0);

	zscale = 1.0 / zscale;
	no_flip_start_az = pref_dir - (90.0 + FLIP_DELAY);
	no_flip_stop_az = pref_dir + (90.0 + FLIP_DELAY);
	start_az = pref_dir - 90.0;
	stop_az = pref_dir + 90.0;
	flip_start_az = pref_dir - (90.0 - FLIP_DELAY);
	flip_stop_az = pref_dir + (90.0 - FLIP_DELAY);

	/* activate map region clip path
	 * This function sets up a clip path so that only plotting
	 * inside the map area will be drawn on paper. GMT_map_setup
	 * must have been called first.  If r >= 0, the map area will
	 * first be painted in the r,g,b colors specified.  flag can
	 * be 0-3, as described in ps_clipon() [in pslib.c].
	 */
	GMT_map_clip_on (GMT_no_rgb, 3);
	
	/* Allocate memory */
		
	/* GMT_memory(a,b,c,d) with a=0 is calloc(b,c) w/error checking */
	/* n_alloc is initialized is GMT_CHUNK = 2000 */
	lon = (double *) GMT_memory (NULL, n_alloc, sizeof (double), "pluswiggle"); /* program name changed 4/13/94 - EAS */
	lat = (double *) GMT_memory (NULL, n_alloc, sizeof (double), "pluswiggle");
	z = (double *) GMT_memory (NULL, n_alloc, sizeof (double), "pluswiggle");
	xx = (double *) GMT_memory (NULL, n_alloc, sizeof (double), "pluswiggle");
	yy = (double *) GMT_memory (NULL, n_alloc, sizeof (double), "pluswiggle");
	zz = (double *) GMT_memory (NULL, n_alloc, sizeof (double), "pluswiggle");
	if (smooth_headings) {
		hh = (double *) GMT_memory (NULL, n_alloc, sizeof (double), "pluswiggle");
		dd = (double *) GMT_memory (NULL, n_alloc, sizeof (double), "pluswiggle");
	}
	
	if (n_files != 0) nofile = FALSE;
	done = FALSE;
	n_args = (argc > 1) ? argc : 2;
	
	for (fno = 1; !done && fno < n_args; fno++) {   /* Loop over all the files */
		if (!nofile && argv[fno][0] == '-') continue;
		if (nofile) {
			fp = stdin;
			done = TRUE;
		}
		else if ((fp = fopen (argv[fno], "r")) == NULL) {
			fprintf (stderr, "pluswiggle: Cannot open file %s\n", argv[fno]);
			continue;
		}
		
		if (!nofile && gmtdefs.verbose) {
			fprintf (stderr, "pluswiggle: Working on file %s\n", argv[fno]);
			printf ("%% file %s\n", argv[fno]);
		}
		
		/* xy_toggle TRUE means read/write I/O as lat/lon instead of lon/lat [default = FALSE] */
		/* FIX: Danger danger DANGER get the point? */
		ix = (gmtdefs.xy_toggle);       iy = 1 - ix;
		
		/* io_header means Input data has header records [default = FALSE] */
		if (gmtdefs.io_header) for (i = 0; i < gmtdefs.n_header_recs; i++) fgets (line, 512, fp);
		if (multi_segments) {
			fgets (line, 512, fp);
			if (gmtdefs.verbose) ps_comment (line);
		}
		more = fgets (line, 512, fp);

		while (more) {
			n = 0;
			while ((more && !multi_segments) || (more && multi_segments && line[0] != EOF_flag)) {
				if (excluded_headings) { /* set the ship_heading_ok flag */
					/* note: heading from gmtlist will be between 0.0 and 360.0 */
					sscanf (line, "%lf %lf %lf %lf", &xy[ix], &xy[iy], &z[n], &ship_heading);
					/* adjust z value immediately */
					if (use_zoffset) z[n] += zoffset;
					/* determine complementary heading angle */
					if (ship_heading >= 180.0) 
						ship_heading_c = ship_heading - 180.0;
					else 
						ship_heading_c = ship_heading + 180.0;
					if (around_zero_degs) /* pick correct rule */
					  ship_heading_ok = ((ship_heading   >= included_headings_min) || 
							     (ship_heading   <= included_headings_max) ||
							     (ship_heading_c >= included_headings_min) ||
							     (ship_heading_c <= included_headings_max) );
					else
					  ship_heading_ok = (((ship_heading >= included_headings_min) &&
							      (ship_heading <= included_headings_max)) ||
							     ((ship_heading_c >= included_headings_min) &&
							      (ship_heading_c <= included_headings_max)));
				}
				else {
					sscanf (line, "%lf %lf %lf", &xy[ix], &xy[iy], &z[n]);
					/* adjust z value immediately */
					if (use_zoffset) z[n] += zoffset;
				}

				if (ship_heading_ok && z_axis_clip) {
					z_axis_ok = (z[n] <= z_axis_max);
					if (z_axis_ok) z_axis_ok = (z[n] >= z_axis_min);
				}
					
				/* this clipping should really be done down below so that GMT_geo_to_xy 
				 * is called only once, but it's simpler & safer to put here - ES */
				if (ship_heading_ok && z_axis_ok && xy_clip) {
					GMT_geo_to_xy (xy[ix], xy[iy], &x_out, &y_out);
					if (xy_clip_ok = (x_out >= project_info.xmin && x_out <= project_info.xmax))
					  xy_clip_ok = (y_out >= project_info.ymin && y_out <= project_info.ymax);
				}

				if (ship_heading_ok && z_axis_ok && xy_clip_ok) {
				    lon[n] = xy[0];     lat[n] = xy[1];
				    n++;
				    if (n >= n_alloc) {
						n_alloc += GMT_CHUNK;
						/* GMT_memory(a,b,c,d) with a!=0 is realloc(a, b*c) w/ error checking */
						/* program name changed 4/13/94 - EAS */
						lon = (double *) GMT_memory ((char *)lon, n_alloc, sizeof (double), "pluswiggle"); 
						lat = (double *) GMT_memory ((char *)lat, n_alloc, sizeof (double), "pluswiggle");
						z   = (double *) GMT_memory ((char *)z,   n_alloc, sizeof (double), "pluswiggle");
						xx  = (double *) GMT_memory ((char *)xx,  n_alloc, sizeof (double), "pluswiggle");
						yy  = (double *) GMT_memory ((char *)yy,  n_alloc, sizeof (double), "pluswiggle");
						zz  = (double *) GMT_memory ((char *)zz,  n_alloc, sizeof (double), "pluswiggle");
						if (smooth_headings) {
							hh = (double *) GMT_memory ((char *)hh, n_alloc, sizeof (double), "pluswiggle");
							dd = (double *) GMT_memory ((char *)dd, n_alloc, sizeof (double), "pluswiggle");
						}
				    }
				}
				more = fgets (line, 512, fp);
			}
					
			/* optionally, smooth out the azimuth used for determining the z-axis orientation */
			if (smooth_headings && !use_fixed_az) { 
				/* this won't be the fastest or most elegant way to do the smoothing,  */
				/* but at least it can be followed */
				GMT_geo_to_xy (lon[0], lat[0], &x2, &y2);
				for (i = 1; i < n; i++) {        
					/* get projected headings */
					x1 = x2; y1 = y2;
					GMT_geo_to_xy(lon[i], lat[i], &x2, &y2);
					dx = x2 - x1; dy = y2 - y1;
					if (dx == 0.0 && dy == 0.0)
						/* hh[i-1] = 0.0; */
						skipped_headings++; /* skip heading until actual heading is obtained */
					else {
						hh[i-1] = d_atan2 (dy, dx) * R2D - 360.0;
						/* note: (d_atan2() * R2D) will always return a number between -180. and +180. */
						/*       so if hh[i] is set here, it will be between -540. and -180.0 */
					   /* fill in headings skipped due to multiple data points at same ship position */
					   while (skipped_headings) hh[i-1-skipped_headings--] = hh[i-1];
					   }
					/* get actual metric distances along track */
					dx = lon[i] - lon[i-1];
					dy = lat[i] - lat[i-1];
					if (MAPPING) { /* MAPPING is TRUE when map projections are used (project_info.projection > 5) */
						dx *= cos (0.5 * (lat[i] + lat[i-1]) * D2R);
						dd[i-1] = hypot (dx, dy) * project_info.M_PR_DEG;
					}
					else
						dd[i-1] = hypot (dx, dy);
				}
				/* handle last point(s) boundary condition for skipped headings */
				skipped_headings++; /* add last point to any others */
				if (skipped_headings < n) {
					last_heading = hh[(n-1)-skipped_headings];
					while (skipped_headings--) hh[(n-1)-skipped_headings] = last_heading;
					}

				/* now we've stored all distances & headings, so smooth them out ... */
				/* we're going to replace each heading with a smoothed heading */
				smooth_cache_hh = (double *) GMT_memory (NULL, smooth_points, sizeof (double), "pluswiggle");
				for (i = 0; (i < smooth_points) && (i < n); i++) /* pre-load cache */
					smooth_cache_hh[i] = hh[i];
				for (i = smooth_points; i < (n-smooth_points); i++) {
					az_smoothing = 0.0;
					az_smoothing_num = 1;
					/* adjustment for previous points */
					dist_tot = 0.0;
					for (k = smooth_points-1, j = i-1; (k >= 0) && ((dist_tot+=dd[j]) <= smooth_distance_max); k--, j--) {
						az_smoothing += angular_diff(smooth_cache_hh[k], hh[i]);
						az_smoothing_num++;
					}
					/* adjustment for subsequent points */
					dist_tot = 0.0;
					for (k = i; (k < i+smooth_points) && ((dist_tot+=dd[k]) <= smooth_distance_max); k++) {
						az_smoothing += angular_diff(hh[k+1], hh[i]);
						az_smoothing_num++;
					}
					/* shift cache */
					for (k = 0; k < smooth_points-1; k++) 
						smooth_cache_hh[k] = smooth_cache_hh[k+1];
					smooth_cache_hh[k] = hh[i];
					/* adjust heading */
					hh[i] += az_smoothing/az_smoothing_num;
				}
			}

			/* Converts lon/lat to x/y using the current projection */
			GMT_geo_to_xy (lon[0], lat[0], &xx[0], &yy[0]);
			zz[0] = z[0];
			paint_wiggle = (do_fill && ((negative && z[0] <= 0.0) || (!negative && z[0] >= 0.0)));
			j = 1;
			/* j  = number of points plotted */
			/* n  = number of points to plot */
			/* ds = distance */

			for (i = 1; i < n; i++) {       /* Convert to inches and get distance increments */
				dx = lon[i] - lon[i-1];
				dy = lat[i] - lat[i-1];
				if (MAPPING) { /* MAPPING is TRUE when map projections are used (project_info.projection > 5) */
					dx *= cos (0.5 * (lat[i] + lat[i-1]) * D2R);
					ds = hypot (dx, dy) * project_info.M_PR_DEG;
				}
				else
					ds = hypot (dx, dy);

				GMT_geo_to_xy (lon[i], lat[i], &x_2, &y_2);

				if (j > 0 && (ds > dist_gap || GMT_is_dnan (z[i]))) {     /* Data gap, plot what we have */
					paint_wiggle = (do_fill && ((negative && zz[j-1] <= 0.0) || (!negative && zz[j-1] >= 0.0)));
					current_flip_state = FLIP_STATE_NO_FLIP;
					plot_wiggle (xx, yy, zz, j, zscale, &fill, &pen_o, &pen_t, paint_wiggle, outline, track, use_fixed_az, fixed_az, delayed_flip);

					j = 0;
				}
				else if (j >= MAX_POINTS) { /* currently defined at 400 (see top of this file) */
					paint_wiggle = (do_fill && ((negative && zz[j-1] <= 0.0) || (!negative && zz[j-1] >= 0.0)));
					plot_wiggle (xx, yy, zz, j, zscale, &fill, &pen_o, &pen_t, paint_wiggle, outline, track, use_fixed_az, fixed_az, delayed_flip);
					xx[0] = xx[j-1];
					yy[0] = yy[j-1];
					zz[0] = zz[j-1];
					if (smooth_headings) {
						dd[0] = dd[j-1];
						hh[0] = hh[j-1];
					}
					j = 1;
				}
				else if (!GMT_is_dnan (z[i-1]) && (z[i]*z[i-1] < 0.0 || z[i] == 0.0)) {   /* Crossed 0, add new point and plot */
					dz = z[i] - z[i-1];
					xx[j] = (dz == 0.0) ? xx[j-1] : xx[j-1] + fabs (z[i-1] / dz) * (x_2 - xx[j-1]);
					yy[j] = (dz == 0.0) ? yy[j-1] : yy[j-1] + fabs (z[i-1] / dz) * (y_2 - yy[j-1]);
					if (smooth_headings) {
						dd[j] = (dz == 0.0) ? dd[j-1] : dd[j-1] + fabs (z[i-1] / dz) * (dd[i] - dd[j-1]);
						hh[j] = (dz == 0.0) ? hh[j-1] : yy[j-1] + fabs (z[i-1] / dz) * (hh[i] - hh[j-1]);
					}
					zz[j] = 0.0;
					j++;
					paint_wiggle = (do_fill && ((negative && zz[j-2] <= 0.0) || (!negative && zz[j-2] >= 0.0)));
					plot_wiggle (xx, yy, zz, j, zscale, &fill, &pen_o, &pen_t, paint_wiggle, outline, track, use_fixed_az, fixed_az, delayed_flip);
					xx[0] = xx[j-1];
					yy[0] = yy[j-1];
					zz[0] = zz[j-1];
					if (smooth_headings) {
						dd[0] = dd[j-1];
						hh[0] = hh[j-1];
					}
					j = 1;
				}
				xx[j] = x_2;
				yy[j] = y_2;
				zz[j] = z[i];
				if (smooth_headings) {
					dd[j] = dd[i];
					hh[j] = hh[i];
				}
				if (!GMT_is_dnan (z[i])) j++;
			}
			
			if (j > 1) {
				paint_wiggle = (do_fill && ((negative && zz[j-1] <= 0.0) || (!negative && zz[j-1] >= 0.0)));
				plot_wiggle (xx, yy, zz, j, zscale, &fill, &pen_o, &pen_t, paint_wiggle, outline, track, use_fixed_az, fixed_az, delayed_flip);
			}
			more = fgets (line, 512, fp);
		}
	}
			
	GMT_map_clip_off();
	if (frame_info.plot) {
		ps_setpaint (gmtdefs.basemap_frame_rgb);
		GMT_map_basemap ();
	}
	ps_setpaint (gmtdefs.background_rgb);
	
	if (project_info.three_D) ps_rotatetrans (z_project.xmin, z_project.ymin, 0.0);
	ps_plotend (gmtdefs.last_page);
	
	free ((char *)lon);
	free ((char *)lat);
	free ((char *)z);
	free ((char *)xx);
	free ((char *)yy);
	free ((char *)zz);
	if (dd != NULL) free ((char *)dd);
	if (hh != NULL) free ((char *)hh);
	if (smooth_cache_hh != NULL) free ((char *)smooth_cache_hh);
	
	GMT_end (argc, argv);
	return 0;
}
	

