/* X_OVER will compute cross-overs between 2 legs (or internal cross-overs
 * if both legs are the same) and write out time,lat,lon,cross-over values,
 * heading_track_1, and heading_track_2 for each crossover. The input file
 * is assumed to be an unformatted, fixed length record file:
 *
 * Rec#1:
 *	YEAR	- (int)		Year the cruise begun (first data point)
 *	NPOINTS	- (int)		Total number of records following rec #1
 *	INFO	- (char)	10 character info field.
 * Rec#2-(NPOINTS+1):
 * 	TIME	- (int)  	Stored as seconds from 00:00:00 on Jan. 1, YEAR.
 *	LAT	- (float)	Stored as degrees * 1.0E6 in the interval <-90.,+90.>
 *	LON	- (float)	Stored as degrees * 1.0E6 in the interval <0.,360.>
 *	GRAV	- (int)		Gravity in mGal * 10.
 *	MAG	- (int)		Magnetic anomaly, in gamma
 *	BATH	- (int)		Bathymetry, in meters
 *
 * The user may choose from three interpolation schemes to calculate the value
 * at the crossover. The default type is linear interpolation, but by specifying
 * -a or -c at the command line, akimas spline or natural cubic spline will be
 * used, respectively. To avoid computing crossovers where the time separation
 * between two consecutive points are large (i.e. at data gaps), the time "
 * must not exceed a certain value. The default value is 15 minuts, but can be
 * changed by using the option -g<minuts>. To find all crossovers regardless of
 * data gaps, specify a large timegap, e.g. -g99999.
 *
 * Programmer:	Paul Wessel
 * Date:	15-JAN-1987
 * Last rev:	18-FEB-1989
 * Last rev:	3-AUG-1989:		Swaps leg1 and leg2 if leg2 is lexically less than leg1
 *
 * 23-JULY-1999: Modify to read gmtplus file format using gmtplusinfo.c 
 * and gmtpluslist.c as a model.  Simply copied gmtplus_mgg.c needed routines into this
 * code. 
 *
 * Restrictions:
 *	Max 100000 points (sum of legs or leg1 if internal)
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/file.h>
#include <unistd.h>
#include "x_system.h"

#include "master_index.h"
#include "gmtplus_mgg.h"

#define NPPS		10
#define NSPB		100
#define NPPB		1000
#define MAX_FLOAT	1.0e38
#define MAX_DOUBLE	1.0e100
#define MAXPOINTS	400000
#define MAXSECT		100
#define MAXBLOCK	100

#define MAX(x,y)	((x) > (y)) ? (x) : (y)
#define MIN(x,y)	((x) < (y)) ? (x) : (y)
#define SWAP(f1,f2) {float tmp; tmp = *f1; *f1 = *f2; *f2 = tmp;}

/*--- New defines for gmtplus format ---*/
#define TWO_ZERO 1 
#define TWO_ONE 2
#define TWO_TWO 4
#define TWO_THREE 8
#define TWO_FOUR 16
#define TWO_FIVE 32
#define TWO_SIX 64
#define TWO_SEVEN 128
#define TWO_EIGHT 256
#define TWO_NINE 512
#define TWO_TEN 1024
#define TWO_ELEVEN 2048
#define TWO_TWELVE 4096
#define TWO_THIRTEEN 8192
#define CNULL ((char *)NULL)
/*-- End new defines --*/


typedef char BOOLEAN;
typedef int HANDLE;

extern BOOLEAN over_lap();
extern BOOLEAN find_cross();

static char *oformat = "%9.5f %9.5f %10.1f %10.1f %9.2f %9.2f %9.2f %8.1f %8.1f %8.1f %5.1f %5.1f\n";
static int ttime[MAXPOINTS];		/* time (in seconds) at each point */
static int maxgap = 900;		/* Largest time separation between two points */
static int npoints[2];			/* no of points pr leg */
static int offset[2];			/* array index where first point if found */
static short int gmt_data[MAXPOINTS][3];	/* grav/mag/bath value at each point */
static int int_mode = 0;		/* Type of interpolation (default = linear) */
static int max_window = 1;		/* No of points on each side of crossover to include
  				   	   when interpolating the cross-over value */
static int leg_no[2];			/* leg_no[1] = 0 if internal */
static float lat[MAXPOINTS];		/* latitude for each point */
static float lon[MAXPOINTS];		/* longitude for each point */
				/* Min/max coordinates for pieces of the legs */
static float grid_section[4][MAXSECT+1][MAXBLOCK][2];
BOOLEAN gmt_flag[3] = {FALSE, FALSE, FALSE};		/* TRUE (for each data type) if it exists */

char *gmt_path[10];	/* Max 10 directories for now */
int n_gmt_paths = 0;	/* Number of these directories */

char *memory();
double NaN;


main(argc, argv)
int argc;
char *argv[]; {
  char legname[2][10];		/* Names of the legs */
  char legfile[100];		/* Filename for leg */
  char info[11];		/* Info-header from input-file */
  char tmp[10];

  int n_pts_pr_blk[2];		/* points pr block for each leg */
  int last_pt[2];		/* no of points in last section */
  int pnt_begin[2];		/* first point to check in current section */
  int pnt_end[2];		/* last point to check in current section */
  int rec, ind, delta, rec_no;	/* misc. counters */
  int pnt[2];

  int nxovers;			/* no of crossovers found */
  int nlegs;			/* 2 if external, 1 if internal */
  int year[2];			/* year of leg */
  int n_pts_pr_sect[2];		/* points pr section for each leg */
  int n_sect_pr_blk[2];		/* sections pr block for each leg */
  int last_blk[2];		/* no of blocks - 1 for each leg */
  int last_sect[2];		/* no of sections - 1 for each leg */
  int blk_begin[2];		/* first block to check for each leg */
  int blk_end[2];		/* last block to check for each leg */
  int sect_begin[2];		/* first sect to check in current block */
  int sect_end[2];		/* last sect to check in current block */
  int latitude, longitude;	/* lat/lon * 1.0E6 from .gmt-file */
  int leg, sct[2], blk[2], sect, ntot = 0;
  int arg, block, mode, i;		/* Misc. counters */
 
  BOOLEAN shift_lon = FALSE;	/* TRUE if areas cross datumline */
  BOOLEAN internal;		/* TRUE if leg1 = leg2 */
  BOOLEAN x_flag;		/* TRUE if crossover is found */
  BOOLEAN error = FALSE;	/* TRUE for invalid arguments */
  BOOLEAN verbose = FALSE;	/* Work in silence */
  BOOLEAN ok;			/* misc. booleans */
  BOOLEAN first = TRUE;		/* Used to print out header before first xover */

  double x_cross, y_cross;	/* Coordinates of crossover */
  double t_old = -1.0e38;	/* Time of previous crossover */
  double t_cross[2];		/* time at crossover along each leg */
  double d_cross[3];		/* Gravity/Magnetic/Bathymetry crossover errors */
  double x_gmb[3];			/* Gravity/Magnetic/Bathymetry mean value at crossover point */

  float xm[2][2];		/* Min/max longitude for area in question */
  float ym[2][2];			/* Same for latitude */
  float x_leg_max[2], x_leg_min[2];	/* Min/max longitudes for leg */
  float y_leg_max[2], y_leg_min[2];	/* Same for latitude */
  float x_blk_max, x_blk_min;		/* Min/max for block */
  float y_blk_max, y_blk_min;
  float x_sect_max, x_sect_min;		/* Min/max for section */
  float y_sect_max, y_sect_min;
  float x_old, y_old;			/* Position of previous crossover */
  float h_cross[2];			/* Heading at crossover along each leg */
  float d_factor[3];			/* Scale factors if applicable */
  float last_lon;			/* Used to check if we cross Greenwich */

  FILE *fp;				/* File pointer */
  
  /*------- Variables for gmt plus format -------------- */
  FILE *hp;				/* Header pointer for plus files */
  char gmtfile[100], gmtheader[100], agency[80], cruise[80];
  int  nfiles,j;
  int n_data[14];
  int hour, minute, second;
  int begin_day, begin_mo, begin_yr, end_day, end_mo, end_yr;
  int greenwich;
  int leg_year, n_records;
  double xmin1, xmax1, xmin2, xmax2, ymin, ymax, xmin, xmax, dlat, dlon, old_x;
  struct GMT *gmt;
  struct COL_N col;
  struct GMT_PLUS_REC record;
  int  mag,grv,top; /*-- variables read --- */
  BOOLEAN	debug_print = FALSE;  /*-- Turn off gmtplusinfo type printout --*/
  /*------- End Variables for gmt plus format ---------- */

  double zip = 0.0;
  NaN = zip / zip;

  d_factor[0] = 0.1;			/* Scale mGal*10 to mGal */
  d_factor[1] = d_factor[2] = 1.0;	/* Leave mag/bath untouched */
  leg = 0;
  for (arg = 1; arg < argc; arg++) {
    if (argv[arg][0] == '-') {
      switch (argv[arg][1]) {
        case 'L':	/* Use linear interpolation */
          int_mode = 0;
          break;
        case 'A':	/* Use akima's interpolation */
          int_mode = 1;
          max_window = 3;
          break;
        case 'C':	/* Use natural cubic spline */
          int_mode = 1; /* Not yet implemented, use Akima */
          max_window = 3;
          break;
        case 'N':	/* No of points to use in interpolation */
          max_window = atoi (&argv[arg][2]);
          break;
        case 'W':	/* Get new timegap */
          maxgap = atoi(&argv[arg][2]) * 60;
          if (maxgap == 0) maxgap = 900;	/* 15 minuts default */
          break;
        case 'G':
          d_factor[0] = atof(&argv[arg][2]);
          if (d_factor[0] == 0.0) d_factor[0] = 0.1;
          break;
        case 'M':
          d_factor[1] = atof(&argv[arg][2]);
          if (d_factor[1] == 0.0) d_factor[1] = 1.0;
          break;
        case 'T':
          d_factor[2] = atof(&argv[arg][2]);
          if (d_factor[2] == 0.0) d_factor[2] = 1.0;
          break;
        case 'V':
          verbose = TRUE;
          break;
        default:
          error = TRUE;
          break;
      }
    }
    else {
      strcpy(legname[leg++],argv[arg]);
      if (leg > 2) error = TRUE;
    }
  }
  if (leg == 1) {
    strcpy(legname[1],legname[0]);
    internal = TRUE;
  }
  else if (leg == 2)
    internal = (!strcmp(legname[0],legname[1])) ? TRUE : FALSE;
  else
    error = TRUE;
  
  if (argc == 1 || error) {
    fprintf(stderr,"x_over - find crossovers between gmt-files\n\n");
    fprintf(stderr,"Usage : x_over leg1 [leg2] [options]\n");
    fprintf(stderr,"options: -L           linear interpolation at crossover [Default]\n");
    fprintf(stderr,"         -A           Quasi hermite (Akima) spline interpolation\n");
    fprintf(stderr,"         -C           Natural cubic spline interpolation\n");
    fprintf(stderr,"	     -N<np>       No of points to use when interpolating. [Default = 6]\n");
    fprintf(stderr,"         -W[minuts]   maximum timegap (in min) allowed at crossover [Default=15]\n");
    fprintf(stderr,"         -G[factor]   multiply gravity values by factor [Default = 0.1]\n");
    fprintf(stderr,"         -M[factor]   multiply magnetic values by factor [Default = 1]\n");
    fprintf(stderr,"         -T[factor]   multiply bathymetric values by factor [Default = 1]\n");
    fprintf(stderr,"         -V           Verbose, report no of crossovers found.\n");
    exit(-1);
  }

  leg_no[0] = 0;
  if (internal) {	/* Same leg -> Internal crossovers */
    leg_no[1] = 0;
    nlegs = 1;
  }
  else {
    leg_no[1] = 1;
    nlegs = 2;
    if (strcmp (legname[0], legname[1]) > 0) {	/* leg2 is lexically less than leg1, swap */
    	strcpy (tmp, legname[0]);
    	strcpy (legname[0], legname[1]);
    	strcpy (legname[1], tmp);
    }
    	
  }

  /*----- New read of gmt plus file format code ------------*/
    gmtpluspath_init();
    read_gmt_master_index();
    nxovers = offset[0] = offset[1] = 0; /*-- Old variable --*/
    for (leg = 0; leg < nlegs; leg++) 
	{	/* Loop over gmt files */
		if (leg == 1) offset[1] = npoints[0];
	        rec_no = offset[leg]+ 1;  /*-- Emulate old read/store logic --*/
		if (argv[leg][0] == '-') continue;
  		if (gmtpluspath_func (gmtfile, gmtheader, legname[leg])) {
   			fprintf(stderr,"x_over : Cannot find leg %s\n", argv[i]);
     			continue;
  		}
		if ((hp = fopen (gmtheader, "r")) == NULL) {
			fprintf(stderr,"x_over : Could not open header file %s\n", gmtheader);
			continue;
		}
		if ((fp = fopen (gmtfile, "r")) == NULL) {
			fprintf(stderr,"x_over : Could not open data file %s\n", gmtfile);
			fclose(hp);
			continue;
		}
		nfiles++;
		/* Read the header file */
		read_gmt_header(hp, &col, &leg_year, &n_records, agency, cruise);
		/* Read first data record */
		read_first_gmt_datarec(fp, &col);
		gmt = gmt_init (leg_year);	/* Initialize gmt_structure */

		/*--- Transfer header data to the old variables ---*/
		year[leg] = leg_year;
		npoints[leg] = n_records;
	
		/* Start reading data from file */
	
		xmin1 = xmin2 = 360.0;
		xmax1 = xmax2 = -360.0;
		ymin = 180.0;
		ymax = -180.0;
		for (j=0;j<14;j++) n_data[j] = 0;
		greenwich = FALSE;
		for (rec = 0; rec < n_records; rec++)
		{
			/*if (fread((char *)(&record), 40, 1, fp) != 1) {*/
			/* 08/31/95 jcm Read to block out bad flagged data */
			if (read_data_record(&record, &col, fp) == -1) {
				fprintf(stderr,"x_over: Error reading data record no %d\n",rec);
				exit(-1);
			}

			/*-- GFAA -- Use this for x_over gravity */
			if(col.gfaa == -1) grv = NODATA;
			else
			  (col_short[col.gfaa]) ? (grv = record.sdata[col.gfaa-6]): (grv = record.ldata[col.gfaa]);
			if(grv == NODATA)
			  {
			    /*-- printf(" gfaa NaN ");-- debug print --*/
			    gmt_data[rec_no][0] = NODATA;
			  }
			else
			  {
			    gmt_data[rec_no][0] = grv;  /*-- Save the gravity data --*/
			    /*-- printf(" gfaa %9.2f",grv*0.1); -- debug print --*/
			  }

			/*-- TCORR -- Use this for x_over bathymetry */
			if(col.tcor == -1)top = NODATA;
			else
			  (col_short[col.tcor]) ? (top = record.sdata[col.tcor-6]): (top = record.ldata[col.tcor]);
			if(top == NODATA)
			  {
			    /*-- printf(" tcor NaN "); -- debug print --*/
			    gmt_data[rec_no][2] = NODATA;
			  }
			else
			  {
			    gmt_data[rec_no][2] = top;  /*-- Save the bathymetry data --*/			  
			    /*-- printf(" tcor %8.1f",top); -- debug print --*/
			  }	  
	  
			/*-- MANM -- Use this for x_over gravity */
			if(col.manm == -1) mag = NODATA;
			else 
			  (col_short[col.manm]) ? (mag = record.sdata[col.manm-6]):(mag = record.ldata[col.manm]);
			if(mag == NODATA)
			  {
			    /*-- printf(" manm NaN "); -- debug print --*/
			    gmt_data[rec_no][1] = NODATA;
			  }
			else 
			  {
			    gmt_data[rec_no][1] = mag;  /*-- Save the magnetic data --*/
			    /*-- printf(" manm %8.1lf ",mag); -- debug print --*/
			  }

			/*-- printf(" rec_no %d\n",rec_no); -- Debug print --*/  

			if (rec == 0)
			  gmt_date(record.ldata[col.tsec],&begin_yr,&begin_mo,&begin_day,&hour,&minute,&second,gmt);

			for (j=0;j<6;j++)
				if (record.ldata[j] != NODATA) n_data[j]++;
			for (j=0;j<8;j++)
				if (record.sdata[j] != NODATA) n_data[j+6]++;

			/*-- Extract location data --*/
			dlon = record.ldata[col.lon] * 0.000001;
			dlat = record.ldata[col.lat] * 0.000001;
			if ((dlon < 0) || (dlon > 360) || (dlat < -90) || (dlat > 90)) {
				fprintf(stderr,"Bad lat/lon: %.5lf, %.5lf Record: %d\n",
					dlat,dlon,rec);
			} else {
				if (dlon < 180.0) {
					xmin1 = MIN (dlon, xmin1);
					xmax1 = MAX (dlon, xmax1);
				}
				else {
					xmin2 = MIN (dlon, xmin2);
					xmax2 = MAX (dlon, xmax2);
				}
				ymin = MIN (dlat, ymin);
				ymax = MAX (dlat, ymax);
				if (rec > 0 && (fabs(old_x-dlon) > 180.0))
				{
					greenwich = TRUE;
				}
				old_x = dlon;
			}
			/*--- Transfer control data to the old variables ---*/			
			n_pts_pr_sect[leg] = NPPS;
			n_sect_pr_blk[leg] = NSPB;
			n_pts_pr_blk[leg]  = NPPB;
			if (npoints[leg]/n_pts_pr_blk[leg] < 1) 
			{
				n_pts_pr_blk[leg] /= 10;
				n_sect_pr_blk[leg] /= 10;
			}
			last_blk[leg] = npoints[leg]/n_pts_pr_blk[leg];
			last_sect[leg] = (npoints[leg]%n_pts_pr_blk[leg])/n_pts_pr_sect[leg];
			last_pt[leg] = npoints[leg] - last_blk[leg]*n_pts_pr_blk[leg]
    					   - last_sect[leg]*n_pts_pr_sect[leg];

			last_lon = lon[offset[leg]];
		        if(rec_no < (offset[leg] + npoints[leg])) 
			{
			  lat[rec_no]   = (float) dlat;
			  lon[rec_no]   = (float) dlon;
			  ttime[rec_no] = record.ldata[col.tsec]; /*-- Time data --*/
			  for (i = 0; i < 3; i++)
			    if (!gmt_flag[i] && gmt_data[rec_no][i] != NODATA) gmt_flag[i] = TRUE; 
			  if (fabs(lon[rec_no] - last_lon) > 180.0 ) shift_lon = TRUE;
			  last_lon = lon[rec_no];
			  rec_no++;
			}	

		}/*-- End for rec loop --*/
		fclose (fp); 
		gmt_date(record.ldata[col.tsec],&end_yr,&end_mo,&end_day,&hour,&minute,&second,gmt);
		free ((char *)gmt);

		if(debug_print) /*-------- Print code from gmtplusinfo ------------*/
		{		
			if (greenwich) {
				xmin = MAX(xmin1,xmin2);
				xmax = MIN(xmax1,xmax2);
			}
			else {
				xmin = MIN(xmin1,xmin2);
				xmax = MAX(xmax1,xmax2);
			}
			if (xmin > xmax) xmin -= 360.0;
	
			printf("Info: %s Year: %d N_recs: %d W: %.5lf E: %.5lf S: %.5lf N: %.5lf Begins: %d %d %d Ends: %d %d %d",
				cruise, leg_year, n_records, xmin, xmax, ymin, ymax,
				begin_day, begin_mo, begin_yr, end_day, end_mo, end_yr);
			for (j=0; j<6; j++) 
			{
				if ((n_data[j] != 0) && (col.tsec != j)
					&& (col.lat != j) && (col.lon != j)
					&& (col.bad_flag != j)
					&& (first_record.ldata[j] != -1)) {
					printf(" %s: %d",gmt_column_name[first_record.ldata[j]], 
						n_data[j]);
				}
			}
			for (j=0; j<8; j++) 
			{
				if ((n_data[j+6] != 0) && (col.tsec != j+6)
					&& (col.lat != j+6) && (col.lon != j+6)
					&& (col.bad_flag != j+6)
					&& (first_record.sdata[j] != -1)) {
					printf(" %s: %d",gmt_column_name[first_record.sdata[j]], 
						n_data[j+6]);
				}
			}
			printf("\n");			
		} 	/*-------- End print code from gmtplusinfo -----------*/

		ntot += npoints[leg];
		if (ntot > MAXPOINTS) {
		fprintf(stderr,"x_over : ntot = %d, must allocate more memory\n",ntot);
		exit(-1);
		}

	} /*-- loop over gmt files ---*/
    /*-- End New read of gmt plus file format code --*/

  /* Devide data into blocks and sections and find min/max corrdinates for each piece */
  rec = 0;
  for (leg = 0; leg < nlegs; leg++) {
    x_leg_max[leg] = x_sect_max = y_leg_max[leg] = y_sect_max = -100000.0;
    x_leg_min[leg] = x_sect_min = y_leg_min[leg] = y_sect_min =  100000.0;
    rec_no = 0;
    for (block = 0; block <= last_blk[leg]; block++) {
      x_blk_max = y_blk_max = -100000.0;
      x_blk_min = y_blk_min =  100000.0;
      sect = 0;
      while (rec_no < npoints[leg] && sect < n_sect_pr_blk[leg]) {
        if (lon[rec] > 180.0 && shift_lon) lon[rec] -= 360.0;
        if (lat[rec] > y_sect_max) y_sect_max = lat[rec];
        if (lat[rec] < y_sect_min) y_sect_min = lat[rec];
        if (lon[rec] > x_sect_max) x_sect_max = lon[rec];
        if (lon[rec] < x_sect_min) x_sect_min = lon[rec];
        if ((rec_no && !(rec_no%n_pts_pr_sect[leg])) || rec_no == (npoints[leg]-1)) {
          if (x_sect_max > x_blk_max) x_blk_max = x_sect_max;
          if (x_sect_min < x_blk_min) x_blk_min = x_sect_min;
          if (y_sect_max > y_blk_max) y_blk_max = y_sect_max;
          if (y_sect_min < y_blk_min) y_blk_min = y_sect_min;
          /* Now, get the min/max coordinates within this section
           * of the ships track */
          grid_section[0][sect][block][leg] = x_sect_min;
          grid_section[1][sect][block][leg] = x_sect_max;
          grid_section[2][sect][block][leg] = y_sect_min;
          grid_section[3][sect][block][leg] = y_sect_max;
          x_sect_min = x_sect_max = lon[rec];
          y_sect_min = y_sect_max = lat[rec];
          sect++;
        }
        rec_no++;
        rec++;
      } /* end of section */
      /* store min/max coordinates for this block */
      grid_section[0][n_sect_pr_blk[leg]][block][leg] = x_blk_min;
      grid_section[1][n_sect_pr_blk[leg]][block][leg] = x_blk_max;
      grid_section[2][n_sect_pr_blk[leg]][block][leg] = y_blk_min;
      grid_section[3][n_sect_pr_blk[leg]][block][leg] = y_blk_max;
      if (x_blk_min < x_leg_min[leg]) x_leg_min[leg] = x_blk_min;
      if (x_blk_max > x_leg_max[leg]) x_leg_max[leg] = x_blk_max;
      if (y_blk_min < y_leg_min[leg]) y_leg_min[leg] = y_blk_min;
      if (y_blk_max > y_leg_max[leg]) y_leg_max[leg] = y_blk_max;
    } /* end of block */
  } /* end of leg */

  /* Here, all data is read and stored in arrays. The min/max coordinates
   * for the legs, blocks, and sections have been found and stored in
   * grid_section[....]. Let's find some cross-overs. First, see if the
   * two legs (if external) have some area in common.
   */
  x_old = lon[0];
  y_old = lat[0];
  for (leg = 0; leg < 2; leg++) {
    blk_begin[leg] = 0;
    blk_end[leg] = last_blk[leg_no[leg]];
    xm[0][leg] = x_leg_min[leg_no[leg]];
    xm[1][leg] = x_leg_max[leg_no[leg]];
    ym[0][leg] = y_leg_min[leg_no[leg]];
    ym[1][leg] = y_leg_max[leg_no[leg]];
    if (xm[0][leg] > xm[1][leg]) xm[0][leg] = xm[1][leg];
    if (ym[0][leg] > ym[1][leg]) ym[0][leg] = ym[1][leg];
  }
  if (internal || over_lap(xm,ym)) {		/* Yes, the legs do overlap, check further */
    xm[0][1] = MAX(xm[0][0], xm[0][1]);
    xm[1][1] = MIN(xm[1][0], xm[1][1]);
    ym[0][1] = MAX(ym[0][0], ym[0][1]);
    ym[1][1] = MIN(ym[1][0], ym[1][1]);
    if ((xm[1][1]-xm[0][1]) > 180.0) {	/* Area crosses both Greenwich and datumline */
      fprintf(stderr,"x_over: Legs %s and %s have more than 180 degrees longitude in common! Aborts\n",
        legname[0], legname[leg_no[1]]);
      exit(-1);
    }
    for (leg = 0; !internal && leg < 2; leg++) {
      for (mode = 0; mode < 2; mode++) {
        ind = (mode) ? blk_begin[leg] : blk_end[leg];
        ok = TRUE;
        while (ok) {
          xm[0][0] = grid_section[0][n_sect_pr_blk[leg]][ind][leg];
          xm[1][0] = grid_section[1][n_sect_pr_blk[leg]][ind][leg];
          ym[0][0] = grid_section[2][n_sect_pr_blk[leg]][ind][leg];
          ym[1][0] = grid_section[3][n_sect_pr_blk[leg]][ind][leg];
          if (over_lap(xm,ym))
            ok = FALSE;
          else if (mode && ind < blk_end[leg])
            ind++;
          else if (!mode && ind > blk_begin[leg])
            ind--;
          else
            ok = FALSE;
        }
        if (mode)
          blk_begin[leg] = ind;
        else
          blk_end[leg] = ind;
      }
    }

    /* - - - - - - - - - - - - - - - -
     *   L E V E L 1 :  B L O C K S
     * - - - - - - - - - - - - - - - -
     * Start comparing two blocks, one from each leg
     */

    for (blk[0] = blk_begin[0]; blk[0] <= blk_end[0]; blk[0]++) {
      for (blk[1] = (internal) ? blk[0] : blk_begin[1]; blk[1] <= blk_end[1]; blk[1]++) {
        /* Now, check if these two blocks overlap */
        for (leg = 0; leg < 2; leg++) {
          xm[0][leg] = grid_section[0][n_sect_pr_blk[leg_no[leg]]][blk[leg]][leg_no[leg]];
          xm[1][leg] = grid_section[1][n_sect_pr_blk[leg_no[leg]]][blk[leg]][leg_no[leg]];
          ym[0][leg] = grid_section[2][n_sect_pr_blk[leg_no[leg]]][blk[leg]][leg_no[leg]];
          ym[1][leg] = grid_section[3][n_sect_pr_blk[leg_no[leg]]][blk[leg]][leg_no[leg]];
          if (xm[0][leg] > xm[1][leg]) xm[0][leg] = xm[1][leg];
          if (ym[0][leg] > ym[1][leg]) ym[0][leg] = ym[1][leg];
        }
        if (over_lap(xm,ym)) {
          /* - - - - - - - - - - - - - - - - - -
           *   L E V E L 2 :  S E C T I O N S
           * - - - - - - - - - - - - - - - - - -
           * Since the blocks have area in common, the next step is to look for
           * overlap between sections within those blocks. For internal crossovers
           * this will always be TRUE, obviously.
           */
          for (leg = 0; leg < 2; leg++) {
            sect_begin[leg] = 0;
            sect_end[leg] = (blk[leg] == last_blk[leg_no[leg]]) ? last_sect[leg_no[leg]] : n_sect_pr_blk[leg_no[leg]] - 1;
          }
          /* Find the area these two blocks have in common, unless
           * we're looking for internal crossovers, where blk1 = blk2
           */
          xm[0][1] = MAX(xm[0][0], xm[0][1]);
          xm[1][1] = MIN(xm[1][0], xm[1][1]);
          ym[0][1] = MAX(ym[0][0], ym[0][1]);
          ym[1][1] = MIN(ym[1][0], ym[1][1]);
          if (!internal) {
            for (leg = 0; leg < 2; leg++) {
              for (mode = 0; mode < 2; mode++) {
        	ind = (mode) ? sect_begin[leg] : sect_end[leg];
        	ok = TRUE;
        	while (ok) {
          	  xm[0][0] = grid_section[0][ind][blk[leg]][leg];
          	  xm[1][0] = grid_section[1][ind][blk[leg]][leg];
          	  ym[0][0] = grid_section[2][ind][blk[leg]][leg];
          	  ym[1][0] = grid_section[3][ind][blk[leg]][leg];
          	  if (over_lap(xm,ym))
            	    ok = FALSE;
          	  else if (mode && ind < sect_end[leg])
              	    ind++;
       	          else if (!mode && ind > sect_begin[leg])
              	    ind--;
              	  else
              	    ok = FALSE;
        	}
        	if (mode)
          	  sect_begin[leg] = ind;
        	else
          	  sect_end[leg] = ind;
     	      }
    	    }
    	  }
    	  /* Start comparing the necessary sections */
    	  for (sct[0] = sect_begin[0]; sct[0] <= sect_end[0]; sct[0]++) {
   	    if (internal && blk[0] == blk[1]) sect_begin[1] = sct[0];
    	    for (sct[1] = sect_begin[1]; sct[1] <= sect_end[1]; sct[1]++) {
    	      for (leg = 0; leg < 2; leg++) {
    	        xm[0][leg] = grid_section[0][sct[leg]][blk[leg]][leg_no[leg]];
    	        xm[1][leg] = grid_section[1][sct[leg]][blk[leg]][leg_no[leg]];
    	        ym[0][leg] = grid_section[2][sct[leg]][blk[leg]][leg_no[leg]];
    	        ym[1][leg] = grid_section[3][sct[leg]][blk[leg]][leg_no[leg]];
          	if (xm[0][leg] > xm[1][leg]) xm[0][leg] = xm[1][leg];	
          	if (ym[0][leg] > ym[1][leg]) ym[0][leg] = ym[1][leg];
              }
              if (over_lap(xm,ym)) {
		/* - - - - - - - - - - - - - -
		 *  L E V E L 3 :  P O I N T S
		 * - - - - - - - - - - - - - -
          	 * Here we have overlap between two sections of the ships
          	 * track. We must now look at pairs of adjacent points to
          	 * see if we indeed have a crossover. Note that if internal
          	 * this test will always be TRUE.
          	 */
	        for (leg = 0; leg < 2; leg++) {
		  if (blk[leg] == last_blk[leg_no[leg]] && sct[leg] == last_sect[leg_no[leg]])
		    pnt_end[leg] = npoints[leg_no[leg]] + offset[leg] - 1;
		  else
		    pnt_end[leg] = blk[leg]*n_pts_pr_blk[leg_no[leg]]
		      + sct[leg]*n_pts_pr_sect[leg_no[leg]] + n_pts_pr_sect[leg_no[leg]]
		      + offset[leg] - 1;
		  pnt_begin[leg] = blk[leg]*n_pts_pr_blk[leg_no[leg]]
		    + sct[leg]*n_pts_pr_sect[leg_no[leg]] + offset[leg];
		}
		xm[0][1] = MAX(xm[0][0],xm[0][1]);
		xm[1][1] = MIN(xm[1][0],xm[1][1]);
		ym[0][1] = MAX(ym[0][0],ym[0][1]);
		ym[1][1] = MIN(ym[1][0],ym[1][1]);
		if (!(blk[0] == blk[1] && sct[0] == sct[1] && internal)) {
		  /* Find the first/last point/sect/block of overlap */
		  for (leg = 0; leg < 2; leg++) {
		    for (mode = 0; mode < 2; mode++) {
		      ind = (mode) ? pnt_begin[leg] : pnt_end[leg];
		      ok = TRUE;
		      while (ok) {
		        xm[0][0] = lon[ind];
		        xm[1][0] = lon[ind+1];
		        ym[0][0] = lat[ind];
		        ym[1][0] = lat[ind+1];
		        if (xm[0][0] > xm[1][0]) SWAP(&xm[0][0],&xm[1][0]);
		        if (ym[0][0] > ym[1][0]) SWAP(&ym[0][0],&ym[1][0]);
		        if (over_lap(xm,ym))
		          ok = FALSE;
		        else if (mode && ind < (pnt_end[leg]-1))
		          ind++;
		        else if (!mode && ind > pnt_begin[leg])
		          ind--;
		        else
		          ok = FALSE;
		      }
		      if (mode)
		        pnt_begin[leg] = ind;
		      else
		        pnt_end[leg] = ind;
		    }
		  }
		}
		/* Start comparing pairs of points from each section */
		for (pnt[0] = pnt_begin[0]; pnt[0] <= pnt_end[0]; pnt[0]++) {
		  if (blk[0] == blk[1] && sct[0] == sct[1] && internal)
		    pnt_begin[1] = pnt[0] + 1;
		  xm[0][0] = lon[pnt[0]];
		  xm[1][0] = lon[pnt[0]+1];
		  ym[0][0] = lat[pnt[0]];
		  ym[1][0] = lat[pnt[0]+1];
		  if (xm[0][0] > xm[1][0]) SWAP(&xm[0][0],&xm[1][0]);
		  if (ym[0][0] > ym[1][0]) SWAP(&ym[0][0],&ym[1][0]);
		  for (pnt[1] = pnt_begin[1]; pnt[1] <= pnt_end[1]; pnt[1]++) {
		    xm[0][1] = lon[pnt[1]];
		    xm[1][1] = lon[pnt[1]+1];
		    ym[0][1] = lat[pnt[1]];
		    ym[1][1] = lat[pnt[1]+1];
		    if (xm[0][1] > xm[1][1]) SWAP(&xm[0][1],&xm[1][1]);
		    if (ym[0][1] > ym[1][1]) SWAP(&ym[0][1],&ym[1][1]);
		    delta = abs(pnt[0] - pnt[1]);
		    if (over_lap(xm,ym) && delta > 1) {
		      /* Here we have found a possible crossover. We interpolate to
		       * find the crossover values for position and time along each
		       * track, and check if the crossover position is inside the
		       * area defined by the 4 points. If so we have a crossover.
		       */
		      x_flag = find_cross(&x_cross,&y_cross,t_cross,d_cross,h_cross,x_gmb,pnt,xm,ym);
		      if (x_flag && !(x_cross == x_old && y_cross == y_old) && t_cross[0] != t_old) {
		        /* We have found a genuine crossover */
		        nxovers++;
		        x_old = x_cross;
		        y_old = y_cross;
		        t_old = t_cross[0];
		        if (x_cross < 0.0) x_cross += 360.;
		        for (i = 0; i < 3; i++) {
		          if (d_cross[i] != NODATA) d_cross[i] *= d_factor[i];
		          if (x_gmb[i] != NODATA) x_gmb[i] *= d_factor[i];
		        }
		        if (first) {
			  /* Print out legnames and start-years */
			  printf("%s %d %s %d\n", legname[0], year[0], legname[leg_no[1]], year[leg_no[1]]);
			  first = FALSE;
			}
		        printf(oformat,y_cross,x_cross,t_cross[0],t_cross[1],d_cross[0],d_cross[1],d_cross[2],
		          x_gmb[0],x_gmb[1],x_gmb[2],h_cross[0],h_cross[1]);
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  if (verbose)
    fprintf(stderr,"x_over: Found %5d cross-overs between %s and %s\n",nxovers,legname[0],legname[1]);
}

BOOLEAN find_cross(xc,yc,tc,dc,hc,xvalues,pt,xx,yy)
double *xc, *yc, dc[], tc[], xvalues[];
float hc[], xx[2][2], yy[2][2];
int pt[]; {
  double delx[2], dely[2], grad[2], xc0, t_next;
  double delt[2], dx[2][3];
  double deg_pr_rad = 57.29577951308232087680;
  double ti[6], di[6];
  int leg, i, j, no_cross, n_int, dtype, i1, i2, n_left, n_right, first, error = 0;
  for (leg = 0; leg < 2; leg++) {	/* Copy pertinent info */
    if (pt[leg] == (npoints[leg_no[leg]] + offset[leg] - 1)) return (FALSE);
    delx[leg] = (double)lon[pt[leg]+1] - (double)lon[pt[leg]];
    dely[leg] = (double)lat[pt[leg]+1] - (double)lat[pt[leg]];
    delt[leg] = (double)ttime[pt[leg]+1] - (double)ttime[pt[leg]];
    grad[leg] = (delx[leg] == 0.0) ? grad[leg] = MAX_DOUBLE : dely[leg]/delx[leg];
  }
  /* grad = MAX_DOUBLE means that line due N-S */
  if (delt[0] > maxgap || delt[1] > maxgap) return (FALSE);	/* Data gap */

  /* Find xcross and ycross */

  if (grad[0] == grad[1])
    *xc = MAX_DOUBLE;
  else if (grad[0] == MAX_DOUBLE)
    *xc = (double) lon[pt[0]];
  else if (grad[1] == MAX_DOUBLE)
    *xc = (double) lon[pt[1]];
  else {
    xc0 = (double)lat[pt[1]] - (double)lat[pt[0]] + grad[0]*(double)lon[pt[0]]
          - grad[1]*(double)lon[pt[1]];
    *xc = xc0/(grad[0]-grad[1]);
  }
  if (grad[0] == grad[1])
    *yc = MAX_DOUBLE;
  else if (grad[0] == MAX_DOUBLE)
    *yc = (double)lat[pt[1]] + grad[1]*(*xc - (double)lon[pt[1]]);
  else
    *yc = (double)lat[pt[0]] + grad[0]*(*xc - (double)lon[pt[0]]);

  /* Check if xcross,ycross is inside the xm,ym area */

  if ((*xc) < xx[0][0] || (*xc) > xx[1][0] || (*xc) < xx[0][1] || (*xc) > xx[1][1]) return (FALSE); /* Outside box */
  if ((*yc) < yy[0][0] || (*yc) > yy[1][0] || (*yc) < yy[0][1] || (*yc) > yy[1][1]) return (FALSE);

  /* if the crossover coincides with one of the input points, we might end up computing this
   * crossover twice. To avoid duplicates of this kind, allow such crossovers to occus only
   * at the second point (timewise)
   */

  if (((*xc) == (double)lon[pt[0]+1] && (*yc) == (double)lat[pt[0]+1]) ||
    ((*xc) == (double)lon[pt[1]+1] && (*yc) == (double)lat[pt[1]+1])) return (FALSE);

  /* Compute time, heading + crossover value */

  for (leg = 0; leg < 2; leg++) {

    /* First, let us compute the crossover time along this leg */

    if (delx[leg] != 0.0)
      tc[leg] = (double)ttime[pt[leg]] + (delt[leg]/delx[leg])*(*xc-(double)lon[pt[leg]]);
    else if (dely[leg] != 0.0)
      tc[leg] = (double)ttime[pt[leg]] + (delt[leg]/dely[leg])*(*yc-(double)lat[pt[leg]]);
    else
      tc[leg] = (double)ttime[pt[leg]];

    /* Now, lets find the data value at the crossover point on this leg */

    /* Here we try to find up to 3 points on each side of the crossover, so
     * that a maximum of 6 points will be used for the computation of spline 
     * coefficients. This must be done for all the 3 data-types.
     */
    dx[leg][0] = dx[leg][1] = dx[leg][2] = NODATA;
    for (dtype = 0; dtype < 3; dtype++) {
      if (!gmt_flag[dtype]) continue;	/* Skip if no such data */
      /* First we look to the 'left' */
      i1 = pt[leg];
      t_next = tc[leg];
      n_left = 0;
      first = -1;
      while (i1 >= offset[leg] && n_left < max_window && (t_next - ttime[i1]) <= maxgap) {
        if (gmt_data[i1][dtype] != NODATA) {
          t_next = ttime[i1];
          if (first < 0) first = i1;
          n_left++;
        }
        i1--;
      }
      if (n_left == 0) continue;
      /* Then we look to the 'right' */
      i2 = pt[leg] + 1;
      n_right = 0;
      t_next = ttime[first];
      while (i2 < (npoints[leg_no[leg]]+offset[leg]) && n_right < max_window && (ttime[i2] - t_next) <= maxgap) {
        if (gmt_data[i2][dtype] != NODATA) {
          t_next = ttime[i2];
          n_right++;
        }
        i2++;
      }
      if (n_right == 0) continue;
      n_int = n_right + n_left;
      for (i = i1+1, j = 0; j < n_int; i++) {
        if (gmt_data[i][dtype] != NODATA) {
          ti[j] = ttime[i];
          di[j] = gmt_data[i][dtype];
          j++;
        }
      }

      error = intpol(ti,di,n_int,1,&tc[leg],&(dx[leg][dtype]),int_mode);	/* Do the interpolation */

      if (error != 0) {	/* Oh shit, what could this mean... */
        fprintf(stderr,"x_over : Error = %d returned from intpol\n",error);
        fprintf(stderr,"(pnt[0] = %ld, pnt[1] = %ld\n",pt[0],pt[1]);
        return (FALSE);
      }
    }
    
    /* Finally, lets evaluate the heading at the crossover-point along this leg */

    delx[leg] *= cos(0.5*(lat[pt[leg]+1] + lat[pt[leg]])/deg_pr_rad);
    if (delx[leg] == 0.0)
      hc[leg] = (dely[leg] > 0.0) ? 0.0 : 180.0;
    else {
      hc[leg] = 90 - atan2(dely[leg],delx[leg])*deg_pr_rad;
      if (hc[leg] < 0.0) hc[leg] += 360.;
    }
  }
    
  /* Then compute the cross-over values */
    
  no_cross = 0;
  for (dtype = 0; dtype < 3; dtype++) {
    if (dx[0][dtype] == NODATA || dx[1][dtype] == NODATA) {
      no_cross++;
      dc[dtype] = NODATA;
      xvalues[dtype] = NODATA;
    }
    else {
      xvalues[dtype] = 0.5*(dx[0][dtype] + dx[1][dtype]);
      dc[dtype] = dx[0][dtype] - dx[1][dtype];
    }
  }
  if (no_cross == 3)
    return(FALSE);
  else
    return (TRUE);
}

BOOLEAN over_lap(xx,yy)		/* Checks if the two areas overlap */
float xx[2][2];
float yy[2][2]; {
  if (xx[1][0] < xx[0][1] || xx[0][0] > xx[1][1]) return (FALSE);
  if (yy[1][0] < yy[0][1] || yy[0][0] > yy[1][1]) return (FALSE);
  return (TRUE);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * AKIMA computes the coefficients for a quasi-cubic hermite spline.
 * Same algorithm as in the IMSL library.
 * Programmer:	Paul Wessel
 * Date:	16-JAN-1987
 * Ver:		v.1-pc
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

int akima (x,y,nx,c)
int nx;
double x[], y[], c[]; {
	int i, no;
	double t1, t2, b, rm1, rm2, rm3, rm4;

	if (nx < 4) return (-1);
	for (i = 1; i < nx; i++) if (x[i] <= x[i-1]) return (-2);
  
	rm3 = (y[1] - y[0])/(x[1] - x[0]);
	t1 = rm3 - (y[1] - y[2])/(x[1] - x[2]);
	rm2 = rm3 + t1;
	rm1 = rm2 + t1;
	
	/* get slopes */
	
	no = nx - 2;
	for (i = 0; i < nx; i++) {
		if (i >= no)
			rm4 = rm3 - rm2 + rm3;
		else
			rm4 = (y[i+2] - y[i+1])/(x[i+2] - x[i+1]);
		t1 = fabs(rm4 - rm3);
		t2 = fabs(rm2 - rm1);
		b = t1 + t2;
		c[3*i] = (b != 0.0) ? (t1*rm2 + t2*rm3) / b : 0.5*(rm2 + rm3);
		rm1 = rm2;
		rm2 = rm3;
		rm3 = rm4;
	}
	no = nx - 1;
	
	/* compute the coefficients for the nx-1 intervals */
	
	for (i = 0; i < no; i++) {
		t1 = 1.0 / (x[i+1] - x[i]);
		t2 = (y[i+1] - y[i])*t1;
		b = (c[3*i] + c[3*i+3] - t2 - t2)*t1;
		c[3*i+2] = b*t1;
		c[3*i+1] = -b + (t2 - c[3*i])*t1;
	}
	return (0);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * CSPLINE computes the coefficients for a natural cubic spline.
 * To evaluate, call csplint
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
 
int cspline (x, y, n, c)
double x[], y[], c[];
int n;
{
	int i, k;
	double p, s, *u;

	u = (double *) memory ((char *)NULL, n, sizeof (double), "cspline");
	c[1] = c[n] = u[1] = 0.0;
	for (i = 1; i < n-1; i++) {
		s = (x[i] - x[i-1]) / (x[i+1] - x[i-1]);
		p = s * c[i-1] + 2.0;
		c[i] = (s - 1.0) / p;
		u[i] = (y[i+1] - y[i]) / (x[i+1] - x[i]) - (y[i] - y[i-1]) / (x[i] - x[i-1]);
		u[i] = (6.0 * u[i] / (x[i+1] - x[i-1]) - s * u[i-1]) / p;
	}
	for (k = n-2; k >= 0; k--) c[k] = c[k] * c[k+1] + u[k];
	free ((char *)u);
	
	return (0);
}

double csplint (x, y, c, n, xp, klo)
double x[], y[], c[], xp;
int n, klo;
{
	int khi;
	double h, ih, b, a, yp;

	khi = klo + 1;
	h = x[khi] - x[klo];
	ih = 1.0 / h;
	a = (x[khi] - xp) * ih;
	b = (xp - x[klo]) * ih;
	yp = a * y[klo] + b * y[khi] + ((a*a*a - a) * c[klo] + (b*b*b - b) * c[khi]) * (h*h) / 6.0;
	
	return (yp);
}
 
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * INTPOL will interpolate from the dataset <x,y> onto a new set <u,v>
 * where <x,y> and <u> is supplied by the user. <v> is returned. The 
 * parameter mode governs what interpolation scheme that will be used.
 * If u[i] is outside the range of x, then v[i] will contain NaN.
 *
 * input:  x = x-values of input data
 *	   y = y-values "    "     "
 *	   n = number of data pairs
 *	   m = number of desired interpolated values
 *	   u = x-values of these points
 *	mode = type of interpolation
 *   	  mode = 0 : Linear interpolation
 *   	  mode = 1 : Quasi-cubic hermite spline (akima)
 *   	  mode = 2 : Natural cubic spline (cubspl)
 * output: v = y-values at interpolated points
 *
 * Programmer:	Paul Wessel
 * Date:	16-JAN-1987
 * Ver:		v.2
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

int intpol (x,y,n,m,u,v,mode)
double x[], y[], u[], v[];
int n, m, mode; {
	int i, j, err_flag = 0;
	double dx, *c, csplint ();
	
	if (n < 4) mode = 0;
	if (mode > 0) c = (double *) memory ((char *)NULL, 3*n, sizeof(double), "intpol");

	if (mode == 0) {	/* Linear interpolation */
		for (i = 1; i < n; i++)
		    if (x[i] <= x[i-1]) return (-1);
	}
	else if (mode == 1) 	/* Akimas spline */
		err_flag = akima (x,y,n,c);
	else if (mode == 2)	/* Natural cubic spline */
		err_flag = cspline (x,y,n,c);
	else
		err_flag = -4;
	if (err_flag != 0) {
		free ((char *)c);
		return (err_flag);
	}
	
	/* Compute the interpolated values from the coefficients */
	
	j = 0;
	for (i = 0; i < m; i++) {
		if (u[i] < x[0] || u[i] > x[n-1]) {	/* Desired point outside data range */
			v[i] = NaN;
			continue;
		} 
		while (x[j] > u[i] && j > 0) j--;	/* In case u is not sorted */
		while (j < n && x[j] <= u[i]) j++;
		if (j == n) j--;
		if (j > 0) j--;
		switch (mode) {
			case 0:
				dx = u[i] - x[j];
				v[i] = (y[j+1]-y[j])*dx/(x[j+1]-x[j]) + y[j];
				break;
			case 1:
				dx = u[i] - x[j];
				v[i] = ((c[3*j+2]*dx + c[3*j+1])*dx + c[3*j])*dx + y[j];
				break;
			case 2:
				v[i] = csplint (x, y, c, n, u[i], j);
				break;
		}
	}
	if (mode > 0) free ((char *)c);
	return (0);
}

char *memory (prev_addr, n, size, progname)
char *prev_addr, *progname;
int n, size; {
	char *tmp;

	if (n == 0) return((char *) NULL); /* Take care of n = 0 */
	
	if (prev_addr) {
		if ((tmp = realloc ((char *) prev_addr, (unsigned) (n * size))) == (char *)NULL) {
			fprintf (stderr, "GMT Fatal Error: %s could not reallocate more memory, n = %d\n", progname, n);
			exit (-1);
		}
	}
	else {
		if ((tmp = calloc ((unsigned) n, (unsigned) size)) == (char *)NULL) {
			fprintf (stderr, "GMT Fatal Error: %s could not allocate memory, n = %d\n", progname, n);
			exit (-1);
		}
	}
	return (tmp);
}
	
/*---------------------------------------------------------------
 * Functions to read GMT Plus header and data files from
 * gmtplus_mgg.c and renamed to avoid confusion with old functions.
 *---------------------------------------------------------------
*/

/* gmtpluspath_init reads the LIBDIR/gmtfile_paths file and gets all
 * the gmtfile directories.
 */
 
int gmtpluspath_init () {
	char file[512], line[512];
	FILE *fp;
    char *envdir;

	/* 07/28/94 jcm upgrade to use environment variables */
	/* order: GMTPLUS_LIBDIR, then default LIBDIR */
	envdir = getenv("GMTPLUS_LIBDIR");	
	if (envdir == NULL) {
		sprintf (file, "%s/gmtfile_paths\0", LIBDIR);
	} else {
		strcpy(file,envdir);
		strcat(file,"/gmtfile_paths");
	}
	
	n_gmt_paths = 0;

	if ((fp = fopen (file, "r")) == NULL) {
		fprintf (stderr, "GMTPLUS Warning: path file %s for *.gmt files not found\n", file);
		fprintf (stderr, "(Will only look in current directory for such files)\n");
		return (0);
	}
	
	while (fgets (line, 512, fp)) {
		if (line[0] == '#') continue;	/* Comments */
		if (line[0] == ' ' || line[0] == '\0') continue;	/* Blank line */
		gmt_path[n_gmt_paths] = memory ((char *)NULL, 1, strlen (line), "gmtpath_init");
		/***gmt_path[n_gmt_paths] = GMT_memory (CNULL, 1, strlen (line), "gmtpath_init");***/
		line[strlen (line)-1] = 0;
		strcpy (gmt_path[n_gmt_paths], line);
		n_gmt_paths++;
	}
	fclose (fp);
}



/* gmtpluspath takes a legid as argument and returns the full path
 * to where this data file can be found.  gmtpath_init must be
 * called first
 */
 
int gmtpluspath_func (leg_path, head_path, leg)
char *leg_path, *head_path, *leg; {
	int id;
	char geo_path[512];
	char geo_head[512];
	
	/* First look in current directory */
	sprintf (geo_path, "%s.gmtd\0", leg);
	sprintf (geo_head, "%s.gmth\0", leg);
	if (!access(geo_path, R_OK) && !access(geo_head, R_OK)) {
		strcpy(leg_path, geo_path);
		strcpy(head_path, geo_head);
		return (0);
	}
	
	/* Then look elsewhere */
	
	for (id = 0; id < n_gmt_paths; id++) {
		sprintf (geo_path, "%s/%s.gmtd\0", gmt_path[id], leg);
		sprintf (geo_head, "%s/%s.gmth\0", gmt_path[id], leg);
		if (!access (geo_path, R_OK) && !access(geo_head, R_OK)) {
			strcpy (leg_path, geo_path);
			strcpy (head_path, geo_head);
			return (0);
		}
	}
	return(1);
}

/* GMT function gmt_date computes the date (hr/mi/sec/dd/mm/yy) based
 * on the total time in seconds since the beginning of first_year.
 * The pointer to the GMT structure is passed allong with the other
 * arguments. The Julian day is returned. the yymmddhhmmss is passed
 * through the argument list.
 */
 
int gmt_date(time,year,month,day,hour,minute,second,gmt_struct)
int time, *year, *month, *day, *hour, *minute, *second;
struct GMT *gmt_struct; {
	int day_time, julian_day;
	day_time = time/86400;
	*month = day_time / 31 + 1;
	if (gmt_struct->daymon[*month +1] <= day_time)
	  (*month)++;
	*year = (*month  - 1) / 12 + gmt_struct->first_year;
	*day = day_time - gmt_struct->daymon[*month] + 1;
	julian_day = (*month > 12) ?
		gmt_struct->daymon[*month] - gmt_struct->daymon[(*month - (*month)%12)] + *day :
		gmt_struct->daymon[*month] + *day;
	*month  = (*month-1)%12 + 1;
	time %= 86400;
	*hour = time / 3600;
	*minute = (time%3600) / 60;
	*second = time - *hour * 3600 - *minute * 60;
	return (julian_day);
}

/*
 *	GMT subroutine gmt_init sets up the structure GMT  which is
 *	used by other gmt routines (gmt_time,gmt_date) to convert
 *	times.  Daymon[month] contains the cumulative number of
 *	days from Jan 1 in first_year through the months PRIOR to the
 *	value of month.  0 <= month <= 60.  month = 0 only occurs
 *	during initializing in this routine. The user must declare
 *	a pointer to the struct GMT in the main program and pass it
 *	when calling the gmt_* functions. To define the GMT structure,
 *	include the file gmt.h
 *
 *	Paul Wessel
 *	12-JUL-1987
 *
 */
 
struct GMT *gmt_init(year1)
int year1; {
	struct GMT *gmt_struct;
	int dm[12];	/* No of days in each month */
	int year, this_year, month, m;
	/***gmt_struct = (struct GMT *) GMT_memory (CNULL, 1, sizeof(struct GMT), "gmt_init");***/
	gmt_struct = (struct GMT *) memory (NULL, 1, sizeof(struct GMT), "gmt_init");
	gmt_struct->first_year = year1;
	/* initialize days of the month etc. */
	dm[0] = 0;
	dm[1] = 31;
	dm[2] = 28;
	dm[3] = 31;
	dm[4] = 30;
	dm[5] = 31;
	dm[6] = 30;
	dm[7] = 31;
	dm[8] = 31;
	dm[9] = 30;
	dm[10] = 31;
	dm[11] = 30;
	gmt_struct->daymon[0] = 0;
	for (year = 0, month = 0; year < 5; year++) {
		this_year = gmt_struct->first_year + year;
		if (this_year%4 == 0) dm[2] = 29;
		for (m = 1; m <= 12; m++) {
			month++;
			gmt_struct->daymon[month] = gmt_struct->daymon[month - 1] + dm[m - 1];
		}
   		dm[2] = 28;
   		dm[0] = 31;
   	}
   	return (gmt_struct);
}

/* GMT function gmt_time returns the number of seconds from
 * first_year calculated from (hr/mi/sc/dd/mm/yy). The pointer
 * to the GMT structure is passed along with the arguments.
 */
 /* MODIFIED 10 July, 1987 by W. Smith  --  I killed a bug in
     month calculation */
 
int gmt_time(time,year,month,day,hour,minute,second, gmt_struct)
int *time, year, month, day, hour, minute, second;
struct GMT *gmt_struct; {
	int mon, n_days;
	if ((mon = (year - gmt_struct->first_year)) > 4) {
	  fprintf(stderr,"gmt_time:  Year - first_year > 4\n");
	  return(-1);
	}
	mon = mon * 12 + month;
	n_days = gmt_struct->daymon[mon] + day - 1;
	*time = n_days * 86400 + hour * 3600 + minute * 60 + second;
	return (*time);
}


/* read info from gmt header file & setup column structure */
int read_gmt_header(hp, col, year, nrec, agency, cruise)
	FILE *hp;
	struct COL_N *col;
	int *year;
	int *nrec;
	char *agency;
	char *cruise;
{
	char buffer[256];
	char tag[32];
	char data[80];
	int i;
	int col_data;
	int success;
	char file_name[80];
	


	col->tsec = 0;
	col->lat = 1;
	col->lon = 2;
	col->graw = 3;
	col->gfaa = 6;
	col->twtt = 7;
	col->tcor = 8;
	col->mtot = 9;
	col->manm = 10;
	col->mflt = -1;
	col->bad_flag = -1;
	col->user_long[0] = -1;
	col->user_long[1] = -1;
	col->user_long[2] = -1;
	col->user_long[3] = -1;
	col->user_long[4] = -1;
	col->user_long[5] = -1;
	col->user_short[0] = -1;
	col->user_short[1] = -1;
	col->user_short[2] = -1;
	col->user_short[3] = -1;
	col->user_short[4] = -1;
	col->user_short[5] = -1;
	col->user_short[6] = -1;
	col->user_short[7] = -1;

	sprintf(agency, "None");
	*year = 1994;
	*nrec = 10;

	while (fscanf(hp,"%[^\n]\n",buffer) != EOF) {
		if (!isspace(buffer[0]) && buffer[0] != '#') { /*i.e. not a comment field or blank */

			data[0] = '\0'; /*empty data string before reading in new line */

			sscanf(buffer,"%[^:]:%s",tag,data);
			if (strcmp(tag,"FILE_ID") == 0) {
				strcpy(file_name,data);
			}
			else if (strcmp(tag,"CRUISE") == 0) {
				strcpy(cruise,data);
			}
			else if (strcmp(tag,"NUMBER_OF_RECORDS") == 0) {
				sscanf(data,"%d",nrec);
			}
			else if (strcmp(tag,"AGENCY") == 0) {
				strcpy(agency,data);
			}
			else if (strcmp(tag,"LEG_YEAR") == 0) {
				sscanf(data,"%d",year);
			}
			else {
				success = FALSE;
				for (i=0;i<MAX_TYPES;i++) {
					if (strcmp(tag,gmt_column_name[i]) == 0) {
						success = TRUE;
						sscanf(data,"%d",&col_data);
						switch (i) {
							case TSEC:
								col->tsec = col_data;
								break;
							case LAT:
								col->lat = col_data;
								break;
							case LON:
								col->lon = col_data;
								break;
							case GRAW:
								col->graw = col_data;
								break;
							case TWTT:
								col->twtt = col_data;
								break;
							case TCORR:
								col->tcor = col_data;
								break;
							case GFAA:
								col->gfaa = col_data;
								break;
							case MTOT:
								col->mtot = col_data;
								break;
							case MANM:
								col->manm = col_data;
								break;
							case MFILT:
								col->mflt = col_data;
								break;
							case BAD_FLAG:
								col->bad_flag = col_data;
								break;
							case UL0:
								col->user_long[0] = col_data;
								break;
							case UL1:
								col->user_long[1] = col_data;
								break;
							case UL2:
								col->user_long[2] = col_data;
								break;
							case UL3:
								col->user_long[3] = col_data;
								break;
							case UL4:
								col->user_long[4] = col_data;
								break;
							case UL5:
								col->user_long[5] = col_data;
								break;
							case US0:
								col->user_short[0] = col_data;
								break;
							case US1:
								col->user_short[1] = col_data;
								break;
							case US2:
								col->user_short[2] = col_data;
								break;
							case US3:
								col->user_short[3] = col_data;
								break;
							case US4:
								col->user_short[4] = col_data;
								break;
							case US5:
								col->user_short[5] = col_data;
								break;
							case US6:
								col->user_short[6] = col_data;
								break;
							case US7:
								col->user_short[7] = col_data;
								break;
							default:
								fprintf(stderr,"Warning: invalid line found in header file.\n");
								break;
						}
						i = MAX_TYPES;
					}
				} 
				if (!success)
					fprintf(stderr,"Warning: invalid line found in header file.\n");
			} /* end else-if chain */

		} else { /* Is # or empty line */
			if (buffer[0] == '#') /*fprintf(stdout,"%s\n",buffer);*/
                ;
			else
				fprintf(stderr,"Warning: invalid line in header.\n");
		}
				
	} /* end while */


}

int read_gmt_master_index()
{
	int i;

	for (i=0;i<6;i++)
		col_short[i] = FALSE;
	for (i=6;i<14;i++)
		col_short[i] = TRUE;

}

int read_first_gmt_datarec(fp,col)
	FILE *fp;
	struct COL_N *col;
{
	int i;
	int data;
	if (fread((char *)(&first_record), 40, 1, fp) != 1) {
		fprintf(stderr,"gmtplus: Error reading first data record\n");
		exit(-1);
	}
	
	for (i=0;i<6;i++) {
		data = first_record.ldata[i];
		switch (data) {
			case TSEC:
				col->tsec = i;
				break;
			case LAT:
				col->lat = i;
				break;
			case LON:
				col->lon = i;
				break;
			case GRAW:
				col->graw = i;
				break;
			case TWTT:
				col->twtt = i;
				break;
			case TCORR:
				col->tcor = i;
				break;
			case GFAA:
				col->gfaa = i;
				break;
			case MTOT:
				col->mtot = i;
				break;
			case MANM:
				col->manm = i;
				break;
			case MFILT:
				col->mflt = i;
				break;
			case BAD_FLAG:
				col->bad_flag = i;
				break;
			case UL0:
				col->user_long[0] = i;
				break;
			case UL1:
				col->user_long[1] = i;
				break;
			case UL2:
				col->user_long[2] = i;
				break;
			case UL3:
				col->user_long[3] = i;
				break;
			case UL4:
				col->user_long[4] = i;
				break;
			case UL5:
				col->user_long[5] = i;
				break;
			case US0:
				col->user_short[0] = i;
				break;
			case US1:
				col->user_short[1] = i;
				break;
			case US2:
				col->user_short[2] = i;
				break;
			case US3:
				col->user_short[3] = i;
				break;
			case US4:
				col->user_short[4] = i;
				break;
			case US5:
				col->user_short[5] = i;
				break;
			case US6:
				col->user_short[6] = i;
				break;
			case US7:
				col->user_short[7] = i;
				break;
			case -1:
				break;
			default:
				fprintf(stderr,"read_first_gmt_datarec: invalid data field listed\n");
				break;
		}
	}
	for (i=0;i<8;i++) {
		data = first_record.sdata[i];
		switch (data) {
			case TSEC:
				col->tsec = i+6;
				break;
			case LAT:
				col->lat = i+6;
				break;
			case LON:
				col->lon = i+6;
				break;
			case GRAW:
				col->graw = i+6;
				break;
			case TWTT:
				col->twtt = i+6;
				break;
			case TCORR:
				col->tcor = i+6;
				break;
			case GFAA:
				col->gfaa = i+6;
				break;
			case MTOT:
				col->mtot = i+6;
				break;
			case MANM:
				col->manm = i+6;
				break;
			case MFILT:
				col->mflt = i+6;
				break;
			case BAD_FLAG:
				col->bad_flag = i+6;
				break;
			case UL0:
				col->user_long[0] = i+6;
				break;
			case UL1:
				col->user_long[1] = i+6;
				break;
			case UL2:
				col->user_long[2] = i+6;
				break;
			case UL3:
				col->user_long[3] = i+6;
				break;
			case UL4:
				col->user_long[4] = i+6;
				break;
			case UL5:
				col->user_long[5] = i+6;
				break;
			case US0:
				col->user_short[0] = i+6;
				break;
			case US1:
				col->user_short[1] = i+6;
				break;
			case US2:
				col->user_short[2] = i+6;
				break;
			case US3:
				col->user_short[3] = i+6;
				break;
			case US4:
				col->user_short[4] = i+6;
				break;
			case US5:
				col->user_short[5] = i+6;
				break;
			case US6:
				col->user_short[6] = i+6;
				break;
			case US7:
				col->user_short[7] = i+6;
				break;
			case -1:
				break;
			default:
				fprintf(stderr,"read_first_gmt_datarec: invalid data field listed\n");
				break;
		}
	}
}

int read_data_record(record, col, fp)
struct GMT_PLUS_REC *record;
struct COL_N *col;
FILE *fp;
{
   int flag;

   if (fread((char *)record, 40, 1, fp) != 1) return -1;
   if (col->bad_flag != -1) {
      flag = record->sdata[col->bad_flag];
      if (flag & TWO_ZERO) record->ldata[0] = NODATA;
      if (flag & TWO_ONE) record->ldata[1] = NODATA;
      if (flag & TWO_TWO) record->ldata[2] = NODATA;
      if (flag & TWO_THREE) record->ldata[3] = NODATA;
      if (flag & TWO_FOUR) record->ldata[4] = NODATA;
      if (flag & TWO_FIVE) record->ldata[5] = NODATA;
      if (flag & TWO_SIX) record->sdata[0] = NODATA;
      if (flag & TWO_SEVEN) record->sdata[1] = NODATA;
      if (flag & TWO_EIGHT) record->sdata[2] = NODATA;
      if (flag & TWO_NINE) record->sdata[3] = NODATA;
      if (flag & TWO_TEN) record->sdata[4] = NODATA;
      if (flag & TWO_ELEVEN) record->sdata[5] = NODATA;
      if (flag & TWO_TWELVE) record->sdata[6] = NODATA;
      if (flag & TWO_THIRTEEN) record->sdata[7] = NODATA;
	}
	return 0;
}

int data_is_flagged(flag, col_num)
int flag;
int col_num;
{
   if (flag == 0) return 0;
   switch (col_num) {
      case 0:
         if (flag & TWO_ZERO) return 1; else return 0;
         break;
      case 1:
         if (flag & TWO_ONE) return 1; else return 0;
         break;
      case 2:
         if (flag & TWO_TWO) return 1; else return 0;
         break;
      case 3:
         if (flag & TWO_THREE) return 1; else return 0;
         break;
      case 4:
         if (flag & TWO_FOUR) return 1; else return 0;
         break;
      case 5:
         if (flag & TWO_FIVE) return 1; else return 0;
         break;
      case 6:
         if (flag & TWO_SIX) return 1; else return 0;
         break;
      case 7:
         if (flag & TWO_SEVEN) return 1; else return 0;
         break;
      case 8:
         if (flag & TWO_EIGHT) return 1; else return 0;
         break;
      case 9:
         if (flag & TWO_NINE) return 1; else return 0;
         break;
      case 10:
         if (flag & TWO_TEN) return 1; else return 0;
         break;
      case 11:
         if (flag & TWO_ELEVEN) return 1; else return 0;
         break;
      case 12:
         if (flag & TWO_TWELVE) return 1; else return 0;
         break;
      case 13:
         if (flag & TWO_THIRTEEN) return 1; else return 0;
         break;
   }
   return 1;
}




